from inspect import signature
from typing import Union, Tuple, Optional, Dict

from golem.core.adapter import BaseOptimizationAdapter
from golem.core.adapter.adapter import DomainStructureType
from golem.core.dag.linked_graph import LinkedGraph
from golem.core.optimisers.graph import OptNode, OptGraph
from golem.core.optimisers.graph_builder import GraphBuilder


class OptGraphBuilder(GraphBuilder):
    """ Builder for incremental construction of directed acyclic OptGraphs.
    Semantics:
    - Forward-only & addition-only (can't prepend or delete nodes).
    - Doesn't throw, doesn't fail: methods always have a way to interpret input given current graph state.
    - Is not responsible for the validity of resulting OptGraph (e.g. correct order, valid operations).
    - Builds always new graphs (on copies of nodes), preserves its state between builds. State doesn't leak outside.
    """

    OperationType = Union[str, Tuple[str, dict]]

    def __init__(self, graph_adapter: Optional[BaseOptimizationAdapter] = None, *initial_nodes: Optional[OptNode]):
        super().__init__(graph_adapter, *initial_nodes)

    def add_node(self, operation_type: Optional[str], branch_idx: int = 0, params: Optional[Dict] = None):
        """ Add single node to graph branch of specified index.
        If there are no heads => adds single OptNode.
        If there is single head => adds single OptNode using head as input.
        If there are several heads => adds single OptNode using as input the head indexed by branch_idx.
        If input is None => do nothing.
        If branch_idx is out of bounds => appends new OptNode.

        :param operation_type: new operation, possibly None
        :param branch_idx: index of the head to use as input for the new node
        :param params: parameters dictionary for the specific operation
        :return: self
        """
        if operation_type is None:
            return self
        params = self._pack_params(operation_type, params)
        if branch_idx < len(self.heads):
            input_node = self.heads[branch_idx]
            self.heads[branch_idx] = OptNode(content=params, nodes_from=[input_node])
        else:
            self.heads.append(OptNode(content=params))
        return self

    def add_sequence(self, *operation_type: OperationType, branch_idx: int = 0):
        """ Same as .node() but for many operations at once.

        :param operation_type: operations for new nodes, either as an operation name
            or as a tuple of operation name and operation parameters.
        :param branch_idx: index of the branch for branching its tip
        """
        for operation in operation_type:
            operation, params = self._unpack_operation(operation)
            self.add_node(operation, branch_idx, params)
        return self

    def grow_branches(self, *operation_type: Optional[OperationType]):
        """ Add single node to each branch.

        Argument position means index of the branch to grow.
        None operation means don't grow that branch.
        If there are no nodes => creates new branches.
        If number of input nodes is bigger than number of branches => extra operations create new branches.

        :param operation_type: operations for new nodes, either as an operation name
            or as a tuple of operation name and operation parameters.
        :return: self
        """
        for i, operation in enumerate(operation_type):
            operation, params = self._unpack_operation(operation)
            self.add_node(operation, i, params)
        return self

    def add_branch(self, *operation_type: Optional[OperationType], branch_idx: int = 0):
        """ Create branches at the tip of branch with branch_idx.

        None operations are filtered out.
        Number of new branches equals to number of provided operations.
        If there are no heads => will add several nodes.
        If there is single head => add several nodes using head as the previous.
        If there are several heads => branch head indexed by branch_idx.
        If branch_idx is out of bounds => adds nodes as new heads at the end.
        If no not-None operations are provided, nothing is changed.

        :param operation_type: operations for new nodes, either as an operation name
            or as a tuple of operation name and operation parameters.
        :param branch_idx: index of the branch for branching its tip
        :return: self
        """
        operations = list(filter(None, operation_type))
        if not operations:
            return self
        if branch_idx < len(self.heads):
            input_node = self.heads.pop(branch_idx)
            for i, operation in enumerate(operations):
                operation, params = self._unpack_operation(operation)
                self.heads.insert(branch_idx + i, OptNode(content=self._pack_params(operation, params),
                                                          nodes_from=[input_node]))
        else:
            for operation in operations:
                operation, params = self._unpack_operation(operation)
                self.add_node(operation, self._iend, params)
        return self

    def add_skip_connection_edge(self, branch_idx_first: int, branch_idx_second: int,
                                 node_idx_in_branch_first: int, node_idx_in_branch_second: int):
        """ Joins two nodes which are not placed sequential in one branch.
        Edge is directed from the first node to the second.

        :param branch_idx_first: index of the first branch which to take the first node
        :param branch_idx_second: index of the second branch which to take the second node
        :param node_idx_in_branch_first: index of the node in its branch
        :param node_idx_in_branch_second: index of the node in its branch
        """
        if branch_idx_first >= len(self.heads) or branch_idx_second >= len(self.heads):
            return
        first_node = self._get_node_from_branch_with_idx(branch_idx=branch_idx_first,
                                                         node_idx_in_branch=node_idx_in_branch_first)
        second_node = self._get_node_from_branch_with_idx(branch_idx=branch_idx_second,
                                                          node_idx_in_branch=node_idx_in_branch_second)
        # to avoid cyclic graphs
        if second_node not in first_node.nodes_from and first_node not in second_node.nodes_from:
            second_node.nodes_from.append(first_node)

        return self

    def _get_node_from_branch_with_idx(self, branch_idx: int, node_idx_in_branch: int):
        head_node = self.heads[branch_idx]
        branch_graph = OptGraph(head_node)
        return branch_graph.nodes[node_idx_in_branch]

    def join_branches(self, operation_type: Optional[str], params: Optional[Dict] = None):
        """ Joins all current branches with provided operation as ensemble node.

        If there are no branches => does nothing.
        If there is single branch => adds single node using it as input.
        If there are several branches => adds single node using all heads as inputs.

        :param operation_type: operation to use for joined node
        :param params: parameters dictionary for the specific operation
        :return: self
        """
        if self.heads and operation_type:
            content = self._pack_params(operation_type, params)
            new_head = OptNode(content=content, nodes_from=self.heads)
            self.heads = [new_head]
        return self

    def build(self) -> Optional[DomainStructureType]:
        """ Adapt resulted graph to required graph class. """
        if not self.to_nodes():
            return None
        result_opt_graph = OptGraph(self.to_nodes())
        if self.graph_adapter:
            return self.graph_adapter.restore(result_opt_graph)
        return result_opt_graph

    def merge_with(self, following_builder) -> Optional['OptGraphBuilder']:
        return merge_opt_graph_builders(self, following_builder)


def _merge_adapters(adapter1: Optional[BaseOptimizationAdapter], adapter2: Optional[BaseOptimizationAdapter]) -> \
        Optional[BaseOptimizationAdapter]:
    if adapter1 is None:
        return adapter2
    elif adapter2 is None:
        return adapter1
    merged_dct = {}
    for (k, v1), v2 in zip(vars(adapter1).items(), vars(adapter2).values()):
        if type(v1) != type(v2):
            raise TypeError('adapters have different types of values')
        elif isinstance(v1, bool):
            merged_dct[k] = v1 and v2
        else:
            merged_dct[k] = v2
    true_adapter_type = type(adapter1)
    init_params = signature(true_adapter_type.__init__).parameters
    init_args = {k: merged_dct[k] for k in merged_dct.keys() & init_params}
    merged_adapter = true_adapter_type(**init_args)
    cls_fields = {
        k: merged_dct[k]
        for k in merged_dct.keys() ^ init_params
        if k in merged_dct and k in vars(merged_adapter)}
    vars(merged_adapter).update(cls_fields)
    return merged_adapter


def merge_opt_graph_builders(previous: OptGraphBuilder, following: OptGraphBuilder) -> Optional[OptGraphBuilder]:
    """ Merge two builders.

    Merging is defined for cases one-to-many and many-to-one nodes,
    i.e. one final node to many initial nodes and many final nodes to one initial node.
    Merging is undefined for the case of many-to-many nodes and None is returned.
    Merging of the builder with itself is well-defined and leads to duplication of the graph.

    If one of the builders is empty -- the other one is returned, no merging is performed.
    State of the passed builders is preserved as they were, after merging new builder is returned.

    :return: GraphBuilder if merging is well-defined, None otherwise.
    """

    if not following.heads:
        return previous
    elif not previous.heads:
        return following

    if type(following.graph_adapter) is not type(previous.graph_adapter):
        raise ValueError('Adapters do not match: cannot perform merge')

    merged_graph_adapter = _merge_adapters(previous.graph_adapter, following.graph_adapter)

    lhs_nodes_final = previous.to_nodes()
    rhs_tmp_graph = LinkedGraph(following.to_nodes())
    rhs_nodes_initial = list(filter(lambda node: not node.nodes_from, rhs_tmp_graph.nodes))

    # If merging one-to-one or one-to-many
    if len(lhs_nodes_final) == 1:
        final_node = lhs_nodes_final[0]
        for initial_node in rhs_nodes_initial:
            rhs_tmp_graph.update_node(initial_node,
                                      OptNode(content={'name': initial_node.name}, nodes_from=[final_node]))
    # If merging many-to-one
    elif len(rhs_nodes_initial) == 1:
        initial_node = rhs_nodes_initial[0]
        rhs_tmp_graph.update_node(initial_node,
                                  OptNode(content={'name': initial_node.name}, nodes_from=lhs_nodes_final))
    # Merging is not defined for many-to-many case
    else:
        return None

    # Check that Graph didn't mess up with node types
    if not all(map(lambda n: isinstance(n, OptNode), rhs_tmp_graph.nodes)):
        raise ValueError("Expected Graph only with nodes of type 'OptNode'")
    merged_builder = OptGraphBuilder(merged_graph_adapter, *rhs_tmp_graph.root_nodes())
    return merged_builder



from typing import List

from golem.structural_analysis.graph_sa.results.base_sa_approach_result import BaseSAApproachResult


class DeletionSAApproachResult(BaseSAApproachResult):
    """ Class for presenting deletion result approaches. """

    def __init__(self):
        self.metrics = []

    def add_results(self, metrics_values: List[float]):
        self.metrics = metrics_values

    def get_worst_result(self, metric_idx_to_optimize_by: int) -> float:
        """ Returns the worst metric among all calculated. """
        return self.metrics[metric_idx_to_optimize_by]

    def get_worst_result_with_names(self, metric_idx_to_optimize_by: int) -> dict:
        return {'value': self.get_worst_result(metric_idx_to_optimize_by=metric_idx_to_optimize_by)}

    def get_dict_results(self) -> List[float]:
        """ Returns all calculated results. """
        return self.metrics

    def get_rounded_metrics(self, idx: int = 2) -> list:
        return [round(metric, idx) for metric in self.metrics]

    def __str__(self):
        return 'DeletionSAApproachResult'



import operator
from copy import deepcopy
from functools import reduce
from typing import Sequence, Optional, Any, Tuple, List, Iterable

import numpy as np

from golem.core.dag.graph import Graph
from golem.core.log import default_log
from golem.core.optimisers.adaptive.common_types import TrajectoryStep, GraphTrajectory
from golem.core.optimisers.adaptive.experience_buffer import ExperienceBuffer
from golem.core.optimisers.adaptive.operator_agent import OperatorAgent
from golem.core.optimisers.fitness import Fitness
from golem.core.optimisers.genetic.operators.mutation import Mutation
from golem.core.optimisers.objective import Objective
from golem.core.optimisers.opt_history_objects.individual import Individual
from golem.core.optimisers.opt_history_objects.opt_history import OptHistory
from golem.core.optimisers.opt_history_objects.parent_operator import ParentOperator
from golem.utilities.data_structures import unzip


class AgentTrainer:
    """Utility class providing fit/validate logic for adaptive Mutation agents.
    Works in tandem with `HistoryReader`.

    How to use offline training:

    1. Collect histories to some directory using `ExperimentLauncher`
    2. Create optimizer & Pretrain mutation agent on these histories using `HistoryReader` and `AgentTrainer`
    3. Optionally, validate the Agent on validation set of histories
    4. Run optimization with pretrained agent
    """

    def __init__(self,
                 objective: Objective,
                 mutation_operator: Mutation,
                 agent: Optional[OperatorAgent] = None,
                 ):
        self._log = default_log(self)
        self.agent = agent if agent is not None else mutation_operator.agent
        self.mutation = mutation_operator
        self.objective = objective
        self._adapter = self.mutation.graph_generation_params.adapter

    def fit(self, histories: Iterable[OptHistory], validate_each: int = -1) -> OperatorAgent:
        """
        Method to fit trainer on collected histories.
        param histories: histories to use in training.
        param validate_each: validate agent once in validate_each generation.
        """
        # Set mutation probabilities to 1.0
        initial_req = deepcopy(self.mutation.requirements)
        self.mutation.requirements.mutation_prob = 1.0

        for i, history in enumerate(histories):
            # Preliminary validity check
            # This allows to filter out histories with different objectives automatically
            if history.objective.metric_names != self.objective.metric_names:
                self._log.warning(f'History #{i+1} has different objective! '
                                  f'Expected {self.objective}, got {history.objective}.')
                continue

            # Build datasets
            experience = ExperienceBuffer.from_history(history)
            val_experience = None
            if validate_each > 0 and i % validate_each == 0:
                experience, val_experience = experience.split(ratio=0.8, shuffle=True)

            # Train
            self._log.info(f'Training on history #{i+1} with {len(history.generations)} generations')
            self.agent.partial_fit(experience)

            # Validate
            if val_experience:
                reward_loss, reward_target = self.validate_agent(experience=val_experience)
                self._log.info(f'Agent validation for history #{i+1} & {experience}: '
                               f'Reward target={reward_target:.3f}, loss={reward_loss:.3f}')

        # Reset mutation probabilities to default
        self.mutation.update_requirements(requirements=initial_req)
        return self.agent

    def validate_on_rollouts(self, histories: Sequence[OptHistory]) -> float:
        """Validates rollouts of agent vs. historic trajectories, comparing
        their mean total rewards (i.e. total fitness gain over the trajectory)."""

        # Collect all trajectories from all histories; and their rewards
        trajectories = concat_lists(map(ExperienceBuffer.unroll_trajectories, histories))

        mean_traj_len = int(np.mean([len(tr) for tr in trajectories]))
        traj_rewards = [sum(reward for _, reward, _ in traj) for traj in trajectories]
        mean_baseline_reward = np.mean(traj_rewards)

        # Collect same number of trajectories of the same length; and their rewards
        agent_trajectories = [self._sample_trajectory(initial=tr[0][0], length=mean_traj_len)
                              for tr in trajectories]
        agent_traj_rewards = [sum(reward for _, reward, _ in traj) for traj in agent_trajectories]
        mean_agent_reward = np.mean(agent_traj_rewards)

        # Compute improvement score of agent over baseline histories
        improvement = mean_agent_reward - mean_baseline_reward
        return improvement

    def validate_history(self, history: OptHistory) -> Tuple[float, float]:
        """Validates history of mutated individuals against optimal policy."""
        history_trajectories = ExperienceBuffer.unroll_trajectories(history)
        return self._validate_against_optimal(history_trajectories)

    def validate_agent(self,
                       graphs: Optional[Sequence[Graph]] = None,
                       experience: Optional[ExperienceBuffer] = None) -> Tuple[float, float]:
        """Validates agent policy against optimal policy on given graphs."""
        if experience:
            agent_steps = experience.retrieve_trajectories()
        elif graphs:
            agent_steps = [self._make_action_step(Individual(g)) for g in graphs]
        else:
            self._log.warning('Either graphs or history must not be None for validation!')
            return 0., 0.
        return self._validate_against_optimal(trajectories=[agent_steps])

    def _validate_against_optimal(self, trajectories: Sequence[GraphTrajectory]) -> Tuple[float, float]:
        """Validates a policy trajectories against optimal policy
        that at each step always chooses the best action with max reward."""
        reward_losses = []
        reward_targets = []
        for trajectory in trajectories:
            inds, actions, rewards = unzip(trajectory)
            _, best_actions, best_rewards = self._apply_best_action(inds)
            reward_loss = self._compute_reward_loss(rewards, best_rewards)
            reward_losses.append(reward_loss)
            reward_targets.append(np.mean(best_rewards))
        reward_loss = float(np.mean(reward_losses))
        reward_target = float(np.mean(reward_targets))
        return reward_loss, reward_target

    @staticmethod
    def _compute_reward_loss(rewards, optimal_rewards, normalized=False) -> float:
        """Returns difference (or deviation) from optimal reward.
        When normalized, 0. means actual rewards match optimal rewards completely,
        0.5 means they on average deviate by 50% from optimal rewards,
        and 2.2 means they on average deviate by more than 2 times from optimal reward."""
        reward_losses = np.subtract(optimal_rewards, rewards)  # always positive
        if normalized:
            reward_losses = reward_losses / np.abs(optimal_rewards) \
                if np.count_nonzero(optimal_rewards) == optimal_rewards.size else reward_losses
        means = np.mean(reward_losses)
        return float(means)

    def _apply_best_action(self, inds: Sequence[Individual]) -> TrajectoryStep:
        """Returns greedily optimal mutation for given graph and associated reward."""
        candidates = []
        for ind in inds:
            for mutation_id in self.agent.available_actions:
                try:
                    values = self._apply_action(mutation_id, ind)
                    candidates.append(values)
                except Exception as e:
                    self._log.warning(f'Eval error for mutation <{mutation_id}> '
                                      f'on graph: {ind.graph.descriptive_id}:\n{e}')
                    continue
        best_step = max(candidates, key=lambda step: step[-1])
        return best_step

    def _apply_action(self, action: Any, ind: Individual) -> TrajectoryStep:
        new_graph, applied = self.mutation._adapt_and_apply_mutation(ind.graph, action)
        fitness = self._eval_objective(new_graph) if applied else None
        parent_op = ParentOperator(type_='mutation', operators=applied, parent_individuals=ind)
        new_ind = Individual(new_graph, fitness=fitness, parent_operator=parent_op)

        prev_fitness = ind.fitness or self._eval_objective(ind.graph)
        if prev_fitness and fitness:
            reward = prev_fitness.value - fitness.value
        elif prev_fitness and not fitness:
            reward = -1.
        else:
            reward = 0.
        return new_ind, action, reward

    def _eval_objective(self, graph: Graph) -> Fitness:
        return self._adapter.adapt_func(self.objective)(graph)

    def _make_action_step(self, ind: Individual) -> TrajectoryStep:
        action = self.agent.choose_action(ind.graph)
        return self._apply_action(action, ind)

    def _sample_trajectory(self, initial: Individual, length: int) -> GraphTrajectory:
        trajectory = []
        past_ind = initial
        for i in range(length):
            next_ind, action, reward = self._make_action_step(past_ind)
            trajectory.append((next_ind, action, reward))
            past_ind = next_ind
        return trajectory


def concat_lists(lists: Iterable[List]) -> List:
    return reduce(operator.add, lists, [])



import os
import random
from functools import partial

from typing import Callable

from golem.core.adapter.nx_adapter import BaseNetworkxAdapter
from golem.core.dag.graph_verifier import GraphVerifier
from golem.core.dag.verification_rules import DEFAULT_DAG_RULES
from golem.core.optimisers.graph import OptGraph, OptNode
from golem.core.optimisers.objective import Objective
from golem.core.optimisers.opt_node_factory import DefaultOptNodeFactory
from golem.core.paths import project_root
from golem.metrics.graph_metrics import size_diff
from golem.structural_analysis.graph_sa.graph_structural_analysis import GraphStructuralAnalysis
from golem.structural_analysis.graph_sa.sa_requirements import StructuralAnalysisRequirements


def get_opt_graph() -> OptGraph:
    """ Get diverse OptGraph. """
    node4 = OptNode({'name': 'node4'})
    node5 = OptNode({'name': 'node5'})
    node6 = OptNode({'name': 'node6'})
    node2 = OptNode({'name': 'node2'}, nodes_from=[node4, node5])
    node3 = OptNode({'name': 'node3'}, nodes_from=[node6])
    node1 = OptNode({'name': 'node1'}, nodes_from=[node2, node3])
    return OptGraph(node1)


def quality_custom_metric_1(_: OptGraph) -> float:
    """ Get toy metric for demonstration. """
    metric = -1 * random.randint(80, 100) / 100
    return metric


def quality_custom_metric_2(_: OptGraph) -> float:
    """ Get one more toy metric for demonstration. """
    metric = -1 * random.randint(70, 110) / 100
    return metric


def complexity_metric(graph: OptGraph, adapter: BaseNetworkxAdapter, metric: Callable) -> float:
    """ Adapts specified graph and returns metric calculated. """
    adapted_graph = adapter.restore(graph)
    return metric(adapted_graph)


if __name__ == "__main__":
    opt_graph = get_opt_graph()
    opt_graph.show()

    adapter = BaseNetworkxAdapter()
    objective = Objective(
        quality_metrics={
            'quality_custom_1': quality_custom_metric_1,
        },
        complexity_metrics={
            'graph_size': partial(complexity_metric,
                                  adapter=adapter, metric=partial(size_diff, adapter.restore(opt_graph))),
        },
        is_multi_objective=True
    )

    node_factory = DefaultOptNodeFactory()

    requirements = StructuralAnalysisRequirements(graph_verifier=GraphVerifier(DEFAULT_DAG_RULES),
                                                  main_metric_idx=0,
                                                  seed=1, replacement_number_of_random_operations_nodes=2,
                                                  replacement_number_of_random_operations_edges=2)

    path_to_save = os.path.join(project_root(), 'examples', 'sa')
    os.makedirs(path_to_save, exist_ok=True)
    # structural analysis will optimize given graph if at least one of the metrics was increased.
    sa = GraphStructuralAnalysis(objective=objective, node_factory=node_factory,
                                 requirements=requirements,
                                 path_to_save=path_to_save,
                                 is_visualize_per_iteration=False)

    graph, results = sa.optimize(graph=opt_graph, n_jobs=1, max_iter=3)

    # to show SA results on each iteration
    optimized_graph = GraphStructuralAnalysis.visualize_on_graph(graph=get_opt_graph(), analysis_result=results,
                                                                 metric_idx_to_optimize_by=0,
                                                                 mode="by_iteration",
                                                                 font_size_scale=0.6,
                                                                 save_path=path_to_save)

    graph.show()



from enum import Enum
from typing import Any, Dict, Type


def enum_to_json(obj: Enum) -> Dict[str, Any]:
    return { 'value': obj.value }


def enum_from_json(cls: Type[Enum], json_obj: Dict[str, Any]) -> Enum:
    return cls(json_obj['value'])



from copy import deepcopy
from datetime import timedelta
from functools import partial
from typing import Optional, Tuple, Union, Sequence

import optuna
from optuna import Trial, Study
from optuna.trial import FrozenTrial

from golem.core.adapter import BaseOptimizationAdapter
from golem.core.optimisers.graph import OptGraph
from golem.core.optimisers.objective import ObjectiveFunction
from golem.core.tuning.search_space import SearchSpace, get_node_operation_parameter_label
from golem.core.tuning.tuner_interface import BaseTuner, DomainGraphForTune


class OptunaTuner(BaseTuner):
    def __init__(self, objective_evaluate: ObjectiveFunction,
                 search_space: SearchSpace,
                 adapter: Optional[BaseOptimizationAdapter] = None,
                 iterations: int = 100,
                 early_stopping_rounds: Optional[int] = None,
                 timeout: timedelta = timedelta(minutes=5),
                 n_jobs: int = -1,
                 deviation: float = 0.05,
                 objectives_number: int = 1):
        super().__init__(objective_evaluate,
                         search_space,
                         adapter,
                         iterations,
                         early_stopping_rounds,
                         timeout,
                         n_jobs,
                         deviation)
        self.objectives_number = objectives_number
        self.study = None

    def tune(self, graph: DomainGraphForTune, show_progress: bool = True) -> \
            Union[DomainGraphForTune, Sequence[DomainGraphForTune]]:
        graph = self.adapter.adapt(graph)
        predefined_objective = partial(self.objective, graph=graph)
        is_multi_objective = self.objectives_number > 1

        self.init_check(graph)

        self.study = optuna.create_study(directions=['minimize'] * self.objectives_number)

        init_parameters, has_parameters_to_optimize = self._get_initial_point(graph)
        if not has_parameters_to_optimize:
            self._stop_tuning_with_message(f'Graph {graph.graph_description} has no parameters to optimize')
            tuned_graphs = self.init_graph
        else:
            # Enqueue initial point to try
            if init_parameters:
                self.study.enqueue_trial(init_parameters)

            verbosity_level = optuna.logging.INFO if show_progress else optuna.logging.WARNING
            optuna.logging.set_verbosity(verbosity_level)

            self.study.optimize(predefined_objective,
                                n_trials=self.iterations,
                                n_jobs=self.n_jobs,
                                timeout=self.timeout.seconds,
                                callbacks=[self.early_stopping_callback],
                                show_progress_bar=show_progress)

            if not is_multi_objective:
                best_parameters = self.study.best_trials[0].params
                tuned_graphs = self.set_arg_graph(graph, best_parameters)
                self.was_tuned = True
            else:
                tuned_graphs = []
                for best_trial in self.study.best_trials:
                    best_parameters = best_trial.params
                    tuned_graph = self.set_arg_graph(deepcopy(graph), best_parameters)
                    tuned_graphs.append(tuned_graph)
                    self.was_tuned = True
        final_graphs = self.final_check(tuned_graphs, is_multi_objective)
        final_graphs = self.adapter.restore(final_graphs)
        return final_graphs

    def objective(self, trial: Trial, graph: OptGraph) -> Union[float, Sequence[float, ]]:
        new_parameters = self._get_parameters_from_trial(graph, trial)
        new_graph = BaseTuner.set_arg_graph(graph, new_parameters)
        metric_value = self.get_metric_value(new_graph)
        return metric_value

    def _get_parameters_from_trial(self, graph: OptGraph, trial: Trial) -> dict:
        new_parameters = {}
        for node_id, node in enumerate(graph.nodes):
            operation_name = node.name

            # Get available parameters for operation
            tunable_node_params = self.search_space.parameters_per_operation.get(operation_name, {})

            for parameter_name, parameter_properties in tunable_node_params.items():
                node_op_parameter_name = get_node_operation_parameter_label(node_id, operation_name, parameter_name)

                parameter_type = parameter_properties.get('type')
                sampling_scope = parameter_properties.get('sampling-scope')
                if parameter_type == 'discrete':
                    new_parameters.update({node_op_parameter_name:
                                           trial.suggest_int(node_op_parameter_name, *sampling_scope)})
                elif parameter_type == 'continuous':
                    new_parameters.update({node_op_parameter_name:
                                           trial.suggest_float(node_op_parameter_name, *sampling_scope)})
                elif parameter_type == 'categorical':
                    new_parameters.update({node_op_parameter_name:
                                           trial.suggest_categorical(node_op_parameter_name, *sampling_scope)})
        return new_parameters

    def _get_initial_point(self, graph: OptGraph) -> Tuple[dict, bool]:
        initial_parameters = {}
        has_parameters_to_optimize = False
        for node_id, node in enumerate(graph.nodes):
            operation_name = node.name

            # Get available parameters for operation
            tunable_node_params = self.search_space.parameters_per_operation.get(operation_name)

            if tunable_node_params:
                has_parameters_to_optimize = True
                tunable_initial_params = {get_node_operation_parameter_label(node_id, operation_name, p):
                                          node.parameters[p] for p in node.parameters if p in tunable_node_params}
                if tunable_initial_params:
                    initial_parameters.update(tunable_initial_params)
        return initial_parameters, has_parameters_to_optimize

    def early_stopping_callback(self, study: Study, trial: FrozenTrial):
        if self.early_stopping_rounds is not None:
            current_trial_number = trial.number
            best_trial_number = study.best_trial.number
            should_stop = (current_trial_number - best_trial_number) >= self.early_stopping_rounds
            if should_stop:
                self.log.debug('Early stopping rounds criteria was reached')
                study.stop()



from datetime import datetime, timedelta
from functools import partial
from io import StringIO
from itertools import product
from pathlib import Path
from typing import Sequence, Type, Callable, Optional, List

import networkx as nx
import numpy as np

from examples.synthetic_graph_evolution.generators import generate_labeled_graph, graph_kinds
from examples.synthetic_graph_evolution.utils import draw_graphs_subplots
from golem.core.adapter.nx_adapter import BaseNetworkxAdapter
from golem.core.optimisers.genetic.gp_optimizer import EvoGraphOptimizer
from golem.core.optimisers.optimizer import GraphOptimizer
from golem.metrics.edit_distance import get_edit_dist_metric, matrix_edit_dist
from golem.metrics.graph_metrics import \
    spectral_dist, size_diff, degree_distance_kernel, degree_distance, nxgraph_stats


def get_all_quality_metrics(target_graph):
    quality_metrics = {
        'edit_distance': get_edit_dist_metric(target_graph),
        'matrix_edit_dist': partial(matrix_edit_dist, target_graph),
        'sp_adj': partial(spectral_dist, target_graph, kind='adjacency'),
        'sp_lapl': partial(spectral_dist, target_graph, kind='laplacian'),
        'sp_lapl_norm': partial(spectral_dist, target_graph, kind='laplacian_norm'),
        'graph_size': partial(size_diff, target_graph),
        'degree_dist_mmd': partial(degree_distance_kernel, target_graph),
        'degree_dist': partial(degree_distance, target_graph),
    }
    return quality_metrics


def run_experiments(optimizer_setup: Callable,
                    optimizer_cls: Type[GraphOptimizer] = EvoGraphOptimizer,
                    node_types: Optional[Sequence[str]] = None,
                    graph_names: Sequence[str] = graph_kinds,
                    graph_sizes: Sequence[int] = (30, 100, 300),
                    num_trials: int = 1,
                    trial_timeout: Optional[int] = None,
                    trial_iterations: Optional[int] = None,
                    visualize: bool = False,
                    ):
    log = StringIO()
    if not node_types:
        node_types = ['X']
    for graph_name, num_nodes in product(graph_names, graph_sizes):
        experiment_id = f'Experiment [graph={graph_name} graph_size={num_nodes}]'
        file_name = f'{optimizer_cls.__name__[:-9]}_{graph_name}_n{num_nodes}_iter{trial_iterations}'
        trial_results = []
        for i in range(num_trials):
            start_time = datetime.now()
            print(f'\nTrial #{i} of {experiment_id} started at {start_time}', file=log)

            # Generate random target graph and run the optimizer
            target_graph = generate_labeled_graph(graph_name, num_nodes, node_types)
            target_graph = target_graph.reverse()
            # Run optimizer setup
            optimizer, objective = optimizer_setup(target_graph,
                                                   optimizer_cls=optimizer_cls,
                                                   node_types=node_types,
                                                   timeout=timedelta(minutes=trial_timeout) if trial_timeout else None,
                                                   num_iterations=trial_iterations)
            found_graphs = optimizer.optimise(objective)
            found_graph = found_graphs[0] if isinstance(found_graphs, Sequence) else found_graphs
            history = optimizer.history

            trial_results.extend(history.final_choices)
            found_nx_graph = BaseNetworkxAdapter().restore(found_graph)

            duration = datetime.now() - start_time
            print(f'Trial #{i} finished, spent time: {duration}', file=log)
            print('target graph stats: ', nxgraph_stats(target_graph), file=log)
            print('found graph stats: ', nxgraph_stats(found_nx_graph), file=log)
            if visualize:
                draw_graphs_subplots(target_graph, found_nx_graph,
                                     titles=['Target Graph', 'Found Graph'], show=False)
                diversity_filename = f'./results/diversity_hist_{graph_name}_n{num_nodes}.gif'
                history.show.diversity_population(save_path=diversity_filename)
                history.show.diversity_line(show=False)
                history.show.fitness_line()
            result_dir = Path('results') / file_name
            result_dir.mkdir(parents=True, exist_ok=True)
            history.save(result_dir / f'history_trial_{i}.json')

        # Compute mean & std for metrics of trials
        ff = objective.format_fitness
        trial_metrics = np.array([ind.fitness.values for ind in trial_results])
        trial_metrics_mean = trial_metrics.mean(axis=0)
        trial_metrics_std = trial_metrics.std(axis=0)
        print(f'{experiment_id} finished with metrics:\n'
              f'mean={ff(trial_metrics_mean)}\n'
              f' std={ff(trial_metrics_std)}',
              file=log)
        print(log.getvalue())
    return log.getvalue()


def run_trial(target_graph: nx.DiGraph,
              optimizer_setup: Callable,
              optimizer_cls: Type[GraphOptimizer] = EvoGraphOptimizer,
              timeout: Optional[timedelta] = None,
              num_iterations: Optional[int] = None,
              node_types: Optional[List[str]] = None):
    optimizer, objective = optimizer_setup(target_graph,
                                           optimizer_cls=optimizer_cls,
                                           timeout=timeout,
                                           node_types=node_types,
                                           num_iterations=num_iterations)
    found_graphs = optimizer.optimise(objective)
    found_graph = found_graphs[0] if isinstance(found_graphs, Sequence) else found_graphs
    history = optimizer.history
    return found_graph, history



import random
from abc import abstractmethod, ABC
from random import choice
from typing import Optional, Iterable, List

from golem.core.optimisers.graph import OptNode


class OptNodeFactory(ABC):
    @abstractmethod
    def exchange_node(self,
                      node: OptNode) -> Optional[OptNode]:
        """
        Returns new node based on a current node using information about node and advisor.

        :param node: current node that must be changed.
        """
        pass

    @abstractmethod
    def get_parent_node(self, node: OptNode, **kwargs) -> Optional[OptNode]:
        """
        Returns new parent node for the current node
        based on the content of the current node and using advisor.

        :param node: current node for which a parent node is generated
        """
        pass

    @abstractmethod
    def get_node(self, **kwargs) -> Optional[OptNode]:
        """
        Returns new node based on the requirements for a node.
        """
        pass

    @abstractmethod
    def get_all_available_operations(self) -> List[str]:
        """
        Returns all available models and data operations.
        """
        pass


class DefaultOptNodeFactory(OptNodeFactory):
    """Default node factory that either randomly selects
    one node from the provided lists of available nodes
    or returns a node with a random numeric node name
    in the range from 0 to `num_node_types`."""

    def __init__(self,
                 available_node_types: Optional[Iterable[str]] = None,
                 num_node_types: Optional[int] = None):
        self.available_nodes = tuple(available_node_types) if available_node_types else None
        self._num_node_types = num_node_types or 1000

    def get_all_available_operations(self) -> Optional[List[str]]:
        """
        Returns all available models and data operations.
        """
        return self.available_nodes

    def exchange_node(self, node: OptNode) -> OptNode:
        return self.get_node()

    def get_parent_node(self, node: OptNode, **kwargs) -> OptNode:
        return self.get_node(**kwargs)

    def get_node(self, **kwargs) -> OptNode:
        chosen_node_type = choice(self.available_nodes) \
            if self.available_nodes \
            else random.randint(0, self._num_node_types)
        return OptNode(content={'name': chosen_node_type})



from golem.core.optimisers.archive.generation_keeper import ImprovementWatcher
from golem.core.optimisers.genetic.operators.operator import PopulationT
from golem.core.optimisers.genetic.parameters.parameter import AdaptiveParameter


class AdaptiveGraphDepth(AdaptiveParameter[int]):
    """Adaptive graph depth parameter. Max allowed graph depth changes
     during optimisation depending on observed improvements in the population.
     If there are no improvements, then graph depth is incremented.
     If there is an improvement, then graph depth remains the same.
     Can also play a role of static value if :param adaptive: is False."""

    def __init__(self, improvements: ImprovementWatcher,
                 start_depth: int = 1, max_depth: int = 10,
                 max_stagnation_gens: int = 1,
                 adaptive: bool = True):
        self._improvements = improvements
        self._start_depth = start_depth
        self._max_depth = max_depth
        self._current_depth = start_depth
        self._max_stagnation_gens = max_stagnation_gens
        self._adaptive = adaptive

    @property
    def initial(self) -> int:
        return self._start_depth

    def next(self, population: PopulationT = None) -> int:
        if not self._adaptive:
            return self._max_depth
        if self._current_depth >= self._max_depth:
            return self._current_depth
        if self._improvements.stagnation_iter_count >= self._max_stagnation_gens:
            self._current_depth += 1
        return self._current_depth



from typing import List, Any, TypeVar, Generic

from golem.utilities.data_structures import ComparableEnum as Enum

NodeType = TypeVar('NodeType')


class RemoveType(Enum):
    """Defines allowed kinds of removals in Graph. Used by mutations."""
    forbidden = 'forbidden'
    node_only = 'node_only'
    node_rewire = 'node_rewire'
    with_direct_children = 'with_direct_children'
    with_parents = 'with_parents'


class DefaultChangeAdvisor(Generic[NodeType]):
    """
    Class for advising of graph changes during evolution
    """

    def __init__(self, task=None):
        self.task = task

    def propose_change(self, node: NodeType, possible_operations: List[Any]) -> List[Any]:
        return possible_operations

    def can_be_removed(self, node: NodeType) -> RemoveType:
        return RemoveType.node_rewire

    def propose_parent(self, node: NodeType, possible_operations: List[Any]) -> List[Any]:
        return possible_operations



import os
from pathlib import Path
from statistics import mean, stdev
from typing import Any, Dict, List, Optional, Union, Sequence

import numpy as np
from matplotlib import pyplot as plt

from golem.core.log import default_log
from golem.core.optimisers.opt_history_objects.opt_history import OptHistory
from golem.utilities.data_structures import ensure_wrapped_in_sequence
from golem.visualisation.opt_history.arg_constraint_wrapper import ArgConstraintWrapper
from golem.visualisation.opt_history.fitness_line import setup_fitness_plot
from golem.visualisation.opt_history.utils import show_or_save_figure


class MultipleFitnessLines(metaclass=ArgConstraintWrapper):
    """ Class to compare fitness changes during optimization process.
    :param historical_fitnesses: dictionary with labels to display as keys and list of fintess values as dict values."""

    def __init__(self,
                 historical_fitnesses: Dict[str, Sequence[Sequence[Union[float, Sequence[float]]]]],
                 metric_names: List[str],
                 visuals_params: Dict[str, Any] = None):
        self.historical_fitnesses = historical_fitnesses
        self.metric_names = metric_names
        self.visuals_params = visuals_params or {}
        self.log = default_log(self)

    @staticmethod
    def from_saved_histories(experiment_folders: List[str], root_path: os.PathLike) -> 'MultipleFitnessLines':
        """ Loads histories from specified folders extracting only fitness values
         to not store whole histories in memory.
         Args:
            experiment_folders: names of folders with histories for experiment launches.
            root_path: path to the folder with experiments results
        """
        root = Path(root_path)
        historical_fitnesses = {}
        for exp_name in experiment_folders:
            trials = []
            for history_filename in os.listdir(root / exp_name):
                if history_filename.startswith('history'):
                    history = OptHistory.load(root / exp_name / history_filename)
                    trials.append(history.historical_fitness)
                    print(f"Loaded {history_filename}")
            historical_fitnesses[exp_name] = trials
            print(f'Loaded {len(trials)} trial histories for experiment: {exp_name}')
        metric_names = history.objective.metric_names

        return MultipleFitnessLines(historical_fitnesses, metric_names)

    @staticmethod
    def from_histories(histories_to_compare: Dict[str, Sequence['OptHistory']]) -> 'MultipleFitnessLines':
        """
        Args:
            histories_to_compare: dictionary with labels to display as keys and histories as values."""
        historical_fitnesses = {}
        for key, histories in histories_to_compare.items():
            historical_fitnesses.update({key: [history.historical_fitness for history in histories]})
        metric_names = list(histories_to_compare.values())[0][0].objective.metric_names

        return MultipleFitnessLines(historical_fitnesses, metric_names)

    def visualize(self,
                  save_path: Optional[Union[os.PathLike, str]] = None,
                  with_confidence: bool = True,
                  metric_id: int = 0,
                  dpi: Optional[int] = None):
        """ Visualizes the best fitness values during the evolution in the form of line.
        :param save_path: path to save the visualization. If set, then the image will be saved,
            and if not, it will be displayed.
        :param with_confidence: bool param specifying to use confidence interval or not.
        :param metric_id: numeric index of the metric to visualize (for multi-objective opt-n).
        :param dpi: DPI of the output figure.
        """
        save_path = save_path or self.get_predefined_value('save_path')
        dpi = dpi or self.get_predefined_value('dpi')

        fig, ax = plt.subplots(figsize=(6.4, 4.8), facecolor='w')
        xlabel = 'Generation'
        self.plot_multiple_fitness_lines(ax, metric_id, with_confidence)
        setup_fitness_plot(ax, xlabel, title=f'Fitness lines for {self.metric_names[metric_id]}')
        plt.legend()
        show_or_save_figure(fig, save_path, dpi)

    def plot_multiple_fitness_lines(self, ax: plt.axis, metric_id: int = 0, with_confidence: bool = True):
        for histories, label in zip(list(self.historical_fitnesses.values()), list(self.historical_fitnesses.keys())):
            plot_average_fitness_line_per_generations(ax, histories, label,
                                                      with_confidence=with_confidence,
                                                      metric_id=metric_id)

    def get_predefined_value(self, param: str):
        return self.visuals_params.get(param)


def plot_average_fitness_line_per_generations(
        axis: plt.Axes,
        historical_fitnesses: Sequence[Sequence[Union[float, Sequence[float]]]],
        label: Optional[str] = None,
        metric_id: int = 0,
        with_confidence: bool = True,
        z_score: float = 1.96):
    """Plots average fitness line per number of histories
    with confidence interval for given z-score (default z=1.96 is for 95% confidence)."""

    trial_fitnesses: List[List[float]] = []
    for fitnesses in historical_fitnesses:
        best_fitnesses = get_best_fitness_per_generation(fitnesses, metric_id)
        trial_fitnesses.append(best_fitnesses)

    # Get average fitness value with confidence values
    average_fitness_per_gen = []
    confidence_fitness_per_gen = []
    max_generations = max(len(i) for i in trial_fitnesses)
    for i in range(max_generations):
        all_fitness_gen = []
        for fitnesses in trial_fitnesses:
            if i < len(fitnesses):
                all_fitness_gen.append(fitnesses[i])
            else:
                # if history is too short - repeat the best obtained fitness
                all_fitness_gen.append(fitnesses[-1])
        average_fitness_per_gen.append(mean(all_fitness_gen))
        confidence = stdev(all_fitness_gen) / np.sqrt(len(all_fitness_gen)) \
            if len(all_fitness_gen) >= 2 else 0.
        confidence_fitness_per_gen.append(confidence)

    # Compute confidence interval
    xs = np.arange(len(average_fitness_per_gen))
    ys = np.array(average_fitness_per_gen)
    ci = z_score * np.array(confidence_fitness_per_gen)

    axis.plot(xs, average_fitness_per_gen, label=label)
    if with_confidence:
        axis.fill_between(xs, (ys - ci), (ys + ci), alpha=.2)


def get_best_fitness_per_generation(fitnesses: Sequence[Sequence[Union[float, Sequence[float]]]],
                                    metric_id: int = 0,
                                    ) -> List[float]:
    """Per each generation find the best fitness *seen so far*.
    Returns tuple:
    - list of best seen metric up to that generation
    """
    best_metric = np.inf  # Assuming metric minimization
    best_metrics = []

    for gen_num, gen_fitnesses in enumerate(fitnesses[metric_id]):
        target_metric = min(ensure_wrapped_in_sequence(gen_fitnesses))
        if target_metric <= best_metric:
            best_metric = target_metric
        best_metrics.append(best_metric)

    return best_metrics



from pprint import pprint
from typing import List, Sequence, Optional, Dict

import networkx as nx
from matplotlib import pyplot as plt
from sklearn.cluster import KMeans

from examples.synthetic_graph_evolution.utils import draw_graphs_subplots
from examples.adaptive_optimizer.utils import plot_action_values
from golem.core.adapter.nx_adapter import BaseNetworkxAdapter
from golem.core.optimisers.genetic.gp_optimizer import EvoGraphOptimizer
from golem.core.optimisers.genetic.operators.operator import PopulationT
from golem.core.optimisers.objective import Objective


def run_adaptive_mutations(
        optimizer: EvoGraphOptimizer,
        objective: Objective,
        target: Optional[nx.DiGraph] = None,
        visualize: bool = True,
):
    """This experiment setup outputs graphic of relative action probabilities
    for given target/objective and given optimizer setup."""
    stats_action_value_log: List[List[float]] = []

    def log_action_values(next_pop: PopulationT, optimizer: EvoGraphOptimizer):
        values = optimizer.mutation.agent.get_action_values(obs=next_pop[0])
        stats_action_value_log.append(list(values))

    # Setup the logger and run the optimizer
    optimizer.set_iteration_callback(log_action_values)
    found_graphs = optimizer.optimise(objective)
    found_graph = found_graphs[0] if isinstance(found_graphs, Sequence) else found_graphs
    history = optimizer.history
    agent = optimizer.mutation.agent

    print('History of action probabilities:')
    pprint(stats_action_value_log)
    if visualize:
        found_nx_graph = BaseNetworkxAdapter().restore(found_graph)
        final_metrics = objective(found_nx_graph).value
        if target is not None:
            draw_graphs_subplots(target, found_nx_graph,
                                 titles=['Target Graph', f'Found Graph (fitness={final_metrics})'])
        else:
            draw_graphs_subplots(found_nx_graph, titles=[f'Found Graph (fitness={final_metrics})'])
        history.show.fitness_line()
        plot_action_values(stats_action_value_log, action_tags=agent.actions)
        plt.show()
    return stats_action_value_log


def run_adaptive_mutations_with_context(
        optimizer: EvoGraphOptimizer,
        objective: Objective,
        target: Optional[nx.DiGraph] = None,
        visualize: bool = True,
        n_clusters: int = 2
):
    """This experiment setup outputs graphic of relative action probabilities
    for given target/objective and given optimizer setup."""
    stats_action_value_log: Dict[int, List[List[float]]] = dict()
    cluster = KMeans(n_clusters=n_clusters)

    def log_action_values_with_clusters(next_pop: PopulationT, optimizer: EvoGraphOptimizer):
        obs_contexts = optimizer.mutation.agent.get_context(next_pop)
        cluster.fit(obs_contexts)
        centers = cluster.cluster_centers_
        for i, center in enumerate(centers):
            values = optimizer.mutation.agent.get_action_values(obs=center)
            if i not in stats_action_value_log.keys():
                stats_action_value_log[i] = []
            stats_action_value_log[i].append(list(values))

    # Setup the logger and run the optimizer
    optimizer.set_iteration_callback(log_action_values_with_clusters)
    found_graphs = optimizer.optimise(objective)
    found_graph = found_graphs[0] if isinstance(found_graphs, Sequence) else found_graphs
    history = optimizer.history
    agent = optimizer.mutation.agent

    print('History of action probabilities:')
    pprint(stats_action_value_log)
    if visualize:
        found_nx_graph = BaseNetworkxAdapter().restore(found_graph)
        final_metrics = objective(found_nx_graph).value
        if target is not None:
            draw_graphs_subplots(target, found_nx_graph,
                                 titles=['Target Graph', f'Found Graph (fitness={final_metrics})'])
        else:
            draw_graphs_subplots(found_nx_graph, titles=[f'Found Graph (fitness={final_metrics})'])
        history.show.fitness_line()
        for i in range(n_clusters):
            plot_action_values(stats_action_value_log[i], action_tags=agent.actions)
            plt.show()
    return stats_action_value_log



import os
from typing import Optional, Union

import seaborn as sns
from matplotlib import pyplot as plt

from golem.visualisation.opt_history.history_visualization import HistoryVisualization
from golem.visualisation.opt_history.utils import get_history_dataframe, get_description_of_operations_by_tag, \
    show_or_save_figure, TagOperationsMap, LabelsColorMapType


class OperationsKDE(HistoryVisualization):
    def visualize(self, save_path: Optional[Union[os.PathLike, str]] = None, dpi: Optional[int] = None,
                  best_fraction: Optional[float] = None, tags_map: TagOperationsMap = None,
                  palette: Optional[LabelsColorMapType] = None):
        """ Visualizes operations used across generations in the form of KDE.

        :param save_path: path to save the visualization. If set, then the image will be saved, and if not,
            it will be displayed.
        :param dpi: DPI of the output figure.
        :param best_fraction: fraction of the best individuals of each generation that included in the
            visualization. Must be in the interval (0, 1].
        :param tags_map: if specified, all operations in the history are colored and grouped based on the
            provided tags. If None, operations are not grouped.
        :param palette: a map from operation label to its color. If None, colors are picked by fixed colormap
            for every history independently.
        """

        save_path = save_path or self.get_predefined_value('save_path')
        dpi = dpi or self.get_predefined_value('dpi')
        best_fraction = best_fraction or self.get_predefined_value('best_fraction')
        tags_map = tags_map or self.visuals_params.get('tags_map')
        palette = palette or self.visuals_params.get('palette')

        generation_column_name = 'Generation'
        operation_column_name = 'Operation'
        column_for_operation = 'tag' if tags_map else 'node'

        df_history = get_history_dataframe(self.history, best_fraction, tags_map)
        df_history = df_history.rename({'generation': generation_column_name,
                                        column_for_operation: operation_column_name}, axis='columns')
        operations_found = df_history[operation_column_name].unique()
        if tags_map:
            tags_all = list(tags_map.keys())
            operations_found = [t for t in tags_all if t in operations_found]  # Sort and filter.

            nodes_per_tag = df_history.groupby(operation_column_name)['node'].unique()
            legend_per_tag = {tag: get_description_of_operations_by_tag(tag, nodes_per_tag[tag], 22)
                              for tag in operations_found}
            df_history[operation_column_name] = df_history[operation_column_name].map(legend_per_tag)
            operations_found = map(legend_per_tag.get, operations_found)
            if palette:
                palette = {legend_per_tag.get(tag): palette.get(tag) for tag in legend_per_tag}

        if not palette:
            palette = sns.color_palette('tab10', n_colors=len(operations_found))

        plot = sns.displot(
            data=df_history,
            x=generation_column_name,
            hue=operation_column_name,
            hue_order=operations_found,
            kind='kde',
            clip=(0, max(df_history[generation_column_name])),
            multiple='fill',
            palette=palette
        )

        fig = plot.figure
        fig.set_dpi(dpi)
        fig.set_facecolor('w')
        ax = plt.gca()
        ax.set_xticks(range(len(self.history.generations)))
        ax.locator_params(nbins=10)
        str_fraction_of_graphs = 'all' if best_fraction is None else f'top {best_fraction * 100}% of'
        ax.set_ylabel(f'Fraction in {str_fraction_of_graphs} generation graphs')

        show_or_save_figure(fig, save_path, dpi)



from copy import deepcopy
from itertools import chain
from math import ceil
from random import choice, random, sample
from typing import Callable, Union, Iterable, Tuple, TYPE_CHECKING

from golem.core.adapter import register_native
from golem.core.dag.graph_utils import nodes_from_layer, node_depth
from golem.core.optimisers.genetic.gp_operators import equivalent_subtree, replace_subtrees
from golem.core.optimisers.genetic.operators.operator import PopulationT, Operator
from golem.core.optimisers.graph import OptGraph, OptNode
from golem.core.optimisers.opt_history_objects.individual import Individual
from golem.core.optimisers.opt_history_objects.parent_operator import ParentOperator
from golem.core.optimisers.optimization_parameters import GraphRequirements
from golem.core.optimisers.optimizer import GraphGenerationParams
from golem.utilities.data_structures import ComparableEnum as Enum

if TYPE_CHECKING:
    from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters


class CrossoverTypesEnum(Enum):
    subtree = 'subtree'
    one_point = "one_point"
    none = 'none'
    exchange_edges = 'exchange_edges'
    exchange_parents_one = 'exchange_parents_one'
    exchange_parents_both = 'exchange_parents_both'


CrossoverCallable = Callable[[OptGraph, OptGraph, int], Tuple[OptGraph, OptGraph]]


class Crossover(Operator):
    def __init__(self,
                 parameters: 'GPAlgorithmParameters',
                 requirements: GraphRequirements,
                 graph_generation_params: GraphGenerationParams):
        super().__init__(parameters, requirements)
        self.graph_generation_params = graph_generation_params

    def __call__(self, population: PopulationT) -> PopulationT:
        if len(population) == 1:
            new_population = population
        else:
            new_population = []
            for ind_1, ind_2 in Crossover.crossover_parents_selection(population):
                new_population += self._crossover(ind_1, ind_2)
        return new_population

    @staticmethod
    def crossover_parents_selection(population: PopulationT) -> Iterable[Tuple[Individual, Individual]]:
        return zip(population[::2], population[1::2])

    def _crossover(self, ind_first: Individual, ind_second: Individual) -> Tuple[Individual, Individual]:
        crossover_type = choice(self.parameters.crossover_types)

        if self._will_crossover_be_applied(ind_first.graph, ind_second.graph, crossover_type):
            crossover_func = self._get_crossover_function(crossover_type)
            for _ in range(self.parameters.max_num_of_operator_attempts):
                first_object = deepcopy(ind_first.graph)
                second_object = deepcopy(ind_second.graph)
                new_graphs = crossover_func(first_object, second_object, max_depth=self.requirements.max_depth)
                are_correct = all(self.graph_generation_params.verifier(new_graph) for new_graph in new_graphs)
                if are_correct:
                    parent_individuals = (ind_first, ind_second)
                    new_individuals = self._get_individuals(new_graphs, parent_individuals, crossover_type)
                    return new_individuals

            self.log.debug('Number of crossover attempts exceeded. '
                           'Please check optimization parameters for correctness.')

        return ind_first, ind_second

    def _get_crossover_function(self, crossover_type: Union[CrossoverTypesEnum, Callable]) -> Callable:
        if isinstance(crossover_type, Callable):
            crossover_func = crossover_type
        else:
            crossover_func = self._crossover_by_type(crossover_type)
        return self.graph_generation_params.adapter.adapt_func(crossover_func)

    def _crossover_by_type(self, crossover_type: CrossoverTypesEnum) -> CrossoverCallable:
        crossovers = {
            CrossoverTypesEnum.subtree: subtree_crossover,
            CrossoverTypesEnum.one_point: one_point_crossover,
            CrossoverTypesEnum.exchange_edges: exchange_edges_crossover,
            CrossoverTypesEnum.exchange_parents_one: exchange_parents_one_crossover,
            CrossoverTypesEnum.exchange_parents_both: exchange_parents_both_crossover
        }
        if crossover_type in crossovers:
            return crossovers[crossover_type]
        else:
            raise ValueError(f'Required crossover type is not found: {crossover_type}')

    def _get_individuals(self, new_graphs: Tuple[OptGraph, OptGraph], parent_individuals: Tuple[Individual, Individual],
                         crossover_type: Union[CrossoverTypesEnum, Callable]) -> Tuple[Individual, Individual]:
        operator = ParentOperator(type_='crossover',
                                  operators=str(crossover_type),
                                  parent_individuals=parent_individuals)
        metadata = self.requirements.static_individual_metadata
        return tuple(Individual(graph, operator, metadata=metadata) for graph in new_graphs)

    def _will_crossover_be_applied(self, graph_first, graph_second, crossover_type) -> bool:
        return not (graph_first is graph_second or
                    random() > self.parameters.crossover_prob or
                    crossover_type is CrossoverTypesEnum.none)


@register_native
def subtree_crossover(graph_1: OptGraph, graph_2: OptGraph,
                      max_depth: int, inplace: bool = True) -> Tuple[OptGraph, OptGraph]:
    """Performed by the replacement of random subtree
    in first selected parent to random subtree from the second parent"""

    if not inplace:
        graph_1 = deepcopy(graph_1)
        graph_2 = deepcopy(graph_2)
    else:
        graph_1 = graph_1
        graph_2 = graph_2

    random_layer_in_graph_first = choice(range(graph_1.depth))
    min_second_layer = 1 if random_layer_in_graph_first == 0 and graph_2.depth > 1 else 0
    random_layer_in_graph_second = choice(range(min_second_layer, graph_2.depth))

    node_from_graph_first = choice(nodes_from_layer(graph_1, random_layer_in_graph_first))
    node_from_graph_second = choice(nodes_from_layer(graph_2, random_layer_in_graph_second))

    replace_subtrees(graph_1, graph_2, node_from_graph_first, node_from_graph_second,
                     random_layer_in_graph_first, random_layer_in_graph_second, max_depth)

    return graph_1, graph_2


@register_native
def one_point_crossover(graph_first: OptGraph, graph_second: OptGraph, max_depth: int) -> Tuple[OptGraph, OptGraph]:
    """Finds common structural parts between two trees, and after that randomly
    chooses the location of nodes, subtrees of which will be swapped"""
    pairs_of_nodes = equivalent_subtree(graph_first, graph_second)
    if pairs_of_nodes:
        node_from_graph_first, node_from_graph_second = choice(pairs_of_nodes)

        layer_in_graph_first = graph_first.depth - node_depth(node_from_graph_first)
        layer_in_graph_second = graph_second.depth - node_depth(node_from_graph_second)

        replace_subtrees(graph_first, graph_second, node_from_graph_first, node_from_graph_second,
                         layer_in_graph_first, layer_in_graph_second, max_depth)
    return graph_first, graph_second


@register_native
def exchange_edges_crossover(graph_first: OptGraph, graph_second: OptGraph, max_depth):
    """Parents exchange a certain number of edges with each other. The number of
    edges is defined as half of the minimum number of edges of both parents, rounded up"""

    def find_edges_in_other_graph(edges, graph: OptGraph):
        new_edges = []
        for parent, child in edges:
            parent_new = graph.get_nodes_by_name(str(parent))
            if parent_new:
                parent_new = parent_new[0]
            else:
                parent_new = OptNode(str(parent))
                graph.add_node(parent_new)
            child_new = graph.get_nodes_by_name(str(child))
            if child_new:
                child_new = child_new[0]
            else:
                child_new = OptNode(str(child))
                graph.add_node(child_new)
            new_edges.append((parent_new, child_new))
        return new_edges

    edges_1 = graph_first.get_edges()
    edges_2 = graph_second.get_edges()
    count = ceil(min(len(edges_1), len(edges_2)) / 2)
    choice_edges_1 = sample(edges_1, count)
    choice_edges_2 = sample(edges_2, count)

    for parent, child in choice_edges_1:
        child.nodes_from.remove(parent)
    for parent, child in choice_edges_2:
        child.nodes_from.remove(parent)

    old_edges1 = graph_first.get_edges()
    old_edges2 = graph_second.get_edges()

    new_edges_2 = find_edges_in_other_graph(choice_edges_1, graph_second)
    new_edges_1 = find_edges_in_other_graph(choice_edges_2, graph_first)

    for parent, child in new_edges_1:
        if (parent, child) not in old_edges1:
            child.nodes_from.append(parent)
    for parent, child in new_edges_2:
        if (parent, child) not in old_edges2:
            child.nodes_from.append(parent)

    return graph_first, graph_second


@register_native
def exchange_parents_one_crossover(graph_first: OptGraph, graph_second: OptGraph, max_depth: int):
    """For the selected node for the first parent, change the parent nodes to
    the parent nodes of the same node of the second parent. Thus, the first child is obtained.
    The second child is a copy of the second parent"""

    def find_nodes_in_other_graph(nodes, graph: OptGraph):
        new_nodes = []
        for node in nodes:
            new_node = graph.get_nodes_by_name(str(node))
            if new_node:
                new_node = new_node[0]
            else:
                new_node = OptNode(str(node))
                graph.add_node(new_node)
            new_nodes.append(new_node)
        return new_nodes

    edges = graph_second.get_edges()
    nodes_with_parent_or_child = list(set(chain(*edges)))
    if nodes_with_parent_or_child:

        selected_node = choice(nodes_with_parent_or_child)
        parents = selected_node.nodes_from

        node_from_first_graph = find_nodes_in_other_graph([selected_node], graph_first)[0]

        node_from_first_graph.nodes_from = []
        old_edges1 = graph_first.get_edges()

        if parents:
            parents_in_first_graph = find_nodes_in_other_graph(parents, graph_first)
            for parent in parents_in_first_graph:
                if (parent, node_from_first_graph) not in old_edges1:
                    node_from_first_graph.nodes_from.append(parent)

    return graph_first, graph_second


@register_native
def exchange_parents_both_crossover(graph_first: OptGraph, graph_second: OptGraph, max_depth: int):
    """For the selected node for the first parent, change the parent nodes to
    the parent nodes of the same node of the second parent. Thus, the first child is obtained.
    The second child is formed in a similar way"""

    parents_in_first_graph = []
    parents_in_second_graph = []

    def find_nodes_in_other_graph(nodes, graph: OptGraph):
        new_nodes = []
        for node in nodes:
            new_node = graph.get_nodes_by_name(str(node))
            if new_node:
                new_node = new_node[0]
            else:
                new_node = OptNode(str(node))
                graph.add_node(new_node)
            new_nodes.append(new_node)
        return new_nodes

    edges = graph_second.get_edges()
    nodes_with_parent_or_child = list(set(chain(*edges)))
    if nodes_with_parent_or_child:

        selected_node2 = choice(nodes_with_parent_or_child)
        parents2 = selected_node2.nodes_from
        if parents2:
            parents_in_first_graph = find_nodes_in_other_graph(parents2, graph_first)

        selected_node1 = find_nodes_in_other_graph([selected_node2], graph_first)[0]
        parents1 = selected_node1.nodes_from
        if parents1:
            parents_in_second_graph = find_nodes_in_other_graph(parents1, graph_second)

        for p in parents1:
            selected_node1.nodes_from.remove(p)
        for p in parents2:
            selected_node2.nodes_from.remove(p)

        old_edges1 = graph_first.get_edges()
        old_edges2 = graph_second.get_edges()

        for parent in parents_in_first_graph:
            if (parent, selected_node1) not in old_edges1:
                selected_node1.nodes_from.append(parent)

        for parent in parents_in_second_graph:
            if (parent, selected_node2) not in old_edges2:
                selected_node2.nodes_from.append(parent)

    return graph_first, graph_second



from typing import Sequence

from golem.core.optimisers.genetic.gp_optimizer import EvoGraphOptimizer
from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters
from golem.core.optimisers.graph import OptGraph
from golem.core.optimisers.meta.surrogate_evaluator import SurrogateDispatcher
from golem.core.optimisers.meta.surrogate_model import RandomValuesSurrogateModel
from golem.core.optimisers.objective import Objective, ObjectiveFunction
from golem.core.optimisers.optimization_parameters import GraphRequirements
from golem.core.optimisers.optimizer import GraphGenerationParams
from golem.core.optimisers.populational_optimizer import EvaluationAttemptsError, _try_unfit_graph


class SurrogateEachNgenOptimizer(EvoGraphOptimizer):
    """
    Surrogate optimizer that uses surrogate model for evaluating part of individuals

    Additionally, we need to pass surrogate_model object
    """
    def __init__(self,
                 objective: Objective,
                 initial_graphs: Sequence[OptGraph],
                 requirements: GraphRequirements,
                 graph_generation_params: GraphGenerationParams,
                 graph_optimizer_params: GPAlgorithmParameters,
                 surrogate_model=RandomValuesSurrogateModel(),
                 surrogate_each_n_gen=5
                 ):
        super().__init__(objective, initial_graphs, requirements, graph_generation_params, graph_optimizer_params)
        self.surrogate_model = surrogate_model
        self.surrogate_each_n_gen = surrogate_each_n_gen
        self.surrogate_dispatcher = SurrogateDispatcher(adapter=graph_generation_params.adapter,
                                                        n_jobs=requirements.n_jobs,
                                                        graph_cleanup_fn=_try_unfit_graph,
                                                        delegate_evaluator=graph_generation_params.remote_evaluator,
                                                        surrogate_model=surrogate_model)

    def optimise(self, objective: ObjectiveFunction) -> Sequence[OptGraph]:
        # eval_dispatcher defines how to evaluate objective on the whole population
        evaluator = self.eval_dispatcher.dispatch(objective, self.timer)
        # surrogate_dispatcher defines how to evaluate objective with surrogate model
        surrogate_evaluator = self.surrogate_dispatcher.dispatch(objective, self.timer)

        with self.timer, self._progressbar:
            self._initial_population(evaluator)
            while not self.stop_optimization():
                try:
                    if self.generations.generation_num % self.surrogate_each_n_gen == 0:
                        new_population = self._evolve_population(surrogate_evaluator)
                    else:
                        new_population = self._evolve_population(evaluator)
                except EvaluationAttemptsError as ex:
                    self.log.warning(f'Composition process was stopped due to: {ex}')
                    break
                # Adding of new population to history
                self._update_population(new_population)
        self._update_population(self.best_individuals, 'final_choices')
        return [ind.graph for ind in self.best_individuals]



import os.path
from datetime import timedelta
from pathlib import Path
from typing import Type, Optional, Sequence, List, Iterable, Callable, Dict

import numpy as np
from rdkit.Chem import Draw
from rdkit.Chem.rdchem import BondType

from examples.molecule_search.mol_adapter import MolAdapter
from examples.molecule_search.mol_advisor import MolChangeAdvisor
from examples.molecule_search.mol_graph import MolGraph
from examples.molecule_search.mol_graph_parameters import MolGraphRequirements
from examples.molecule_search.mol_mutations import CHEMICAL_MUTATIONS
from examples.molecule_search.mol_metrics import normalized_sa_score, penalised_logp, qed_score, \
    normalized_logp, CLScorer
from golem.core.dag.verification_rules import has_no_self_cycled_nodes, has_no_isolated_components, \
    has_no_isolated_nodes
from golem.core.optimisers.adaptive.agent_trainer import AgentTrainer
from golem.core.optimisers.adaptive.history_collector import HistoryReader
from golem.core.optimisers.adaptive.operator_agent import MutationAgentTypeEnum
from golem.core.optimisers.genetic.gp_optimizer import EvoGraphOptimizer
from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters
from golem.core.optimisers.genetic.operators.crossover import CrossoverTypesEnum
from golem.core.optimisers.genetic.operators.elitism import ElitismTypesEnum
from golem.core.optimisers.genetic.operators.inheritance import GeneticSchemeTypesEnum
from golem.core.optimisers.objective import Objective
from golem.core.optimisers.opt_history_objects.opt_history import OptHistory
from golem.core.optimisers.optimizer import GraphGenerationParams, GraphOptimizer
from golem.core.paths import project_root
from golem.visualisation.opt_history.multiple_fitness_line import MultipleFitnessLines
from golem.visualisation.opt_viz_extra import visualise_pareto


def get_methane() -> MolGraph:
    methane = 'C'
    return MolGraph.from_smiles(methane)


def get_all_mol_metrics() -> Dict[str, Callable]:
    metrics = {'qed_score': qed_score,
               'cl_score': CLScorer(),
               'norm_sa_score': normalized_sa_score,
               'penalised_logp': penalised_logp,
               'norm_log_p': normalized_logp}
    return metrics


def molecule_search_setup(optimizer_cls: Type[GraphOptimizer] = EvoGraphOptimizer,
                          adaptive_kind: MutationAgentTypeEnum = MutationAgentTypeEnum.random,
                          max_heavy_atoms: int = 50,
                          atom_types: Optional[List[str]] = None,
                          bond_types: Sequence[BondType] = (BondType.SINGLE, BondType.DOUBLE, BondType.TRIPLE),
                          timeout: Optional[timedelta] = None,
                          num_iterations: Optional[int] = None,
                          pop_size: int = 20,
                          metrics: Optional[List[str]] = None,
                          initial_molecules: Optional[Sequence[MolGraph]] = None):
    requirements = MolGraphRequirements(
        max_heavy_atoms=max_heavy_atoms,
        available_atom_types=atom_types or ['C', 'N', 'O', 'F', 'P', 'S', 'Cl', 'Br'],
        bond_types=bond_types,
        early_stopping_timeout=np.inf,
        early_stopping_iterations=np.inf,
        keep_n_best=4,
        timeout=timeout,
        num_of_generations=num_iterations,
        keep_history=True,
        n_jobs=1,
        history_dir=None,
    )
    gp_params = GPAlgorithmParameters(
        pop_size=pop_size,
        max_pop_size=pop_size,
        multi_objective=True,
        genetic_scheme_type=GeneticSchemeTypesEnum.steady_state,
        elitism_type=ElitismTypesEnum.replace_worst,
        mutation_types=CHEMICAL_MUTATIONS,
        crossover_types=[CrossoverTypesEnum.none],
        adaptive_mutation_type=adaptive_kind,
    )
    graph_gen_params = GraphGenerationParams(
        adapter=MolAdapter(),
        rules_for_constraint=[has_no_self_cycled_nodes, has_no_isolated_components, has_no_isolated_nodes],
        advisor=MolChangeAdvisor(),
    )

    metrics = metrics or ['qed_score']
    all_metrics = get_all_mol_metrics()
    objective = Objective(
        quality_metrics={metric_name: all_metrics[metric_name] for metric_name in metrics},
        is_multi_objective=len(metrics) > 1
    )

    initial_graphs = initial_molecules or [get_methane()]
    initial_graphs = graph_gen_params.adapter.adapt(initial_graphs)

    # Build the optimizer
    optimiser = optimizer_cls(objective, initial_graphs, requirements, graph_gen_params, gp_params)
    return optimiser, objective


def visualize_results(molecules: Iterable[MolGraph],
                      objective: Objective,
                      history: OptHistory,
                      save_path: Path,
                      show: bool = False):
    save_path.mkdir(parents=True, exist_ok=True)

    # Plot pareto front (if multi-objective)
    if objective.is_multi_objective:
        visualise_pareto(history.archive_history[-1],
                         objectives_names=objective.metric_names[:2],
                         folder=str(save_path))

    # Plot fitness convergence
    history.show.fitness_line(dpi=100, save_path=save_path / 'fitness_line.png')
    # Plot diversity
    history.show.diversity_population(save_path=save_path / 'diversity.gif')
    history.show.diversity_line(save_path=save_path / 'diversity_line.png')

    # Plot found molecules
    rw_molecules = [mol.get_rw_molecule() for mol in set(molecules)]
    objectives = [objective.format_fitness(objective(mol)) for mol in set(molecules)]
    image = Draw.MolsToGridImage(rw_molecules,
                                 legends=objectives,
                                 molsPerRow=min(4, len(rw_molecules)),
                                 subImgSize=(1000, 1000),
                                 legendFontSize=50)
    image.save(save_path / 'best_molecules.png')
    if show:
        image.show()


def pretrain_agent(optimizer: EvoGraphOptimizer, objective: Objective, results_dir: str) -> AgentTrainer:
    agent = optimizer.mutation.agent
    trainer = AgentTrainer(objective, optimizer.mutation, agent)
    # load histories
    history_reader = HistoryReader(Path(results_dir))
    # train agent
    trainer.fit(histories=history_reader.load_histories(), validate_each=1)
    return trainer


def run_experiment(optimizer_setup: Callable,
                   optimizer_cls: Type[GraphOptimizer] = EvoGraphOptimizer,
                   adaptive_kind: MutationAgentTypeEnum = MutationAgentTypeEnum.random,
                   max_heavy_atoms: int = 50,
                   atom_types: Optional[List[str]] = None,
                   bond_types: Sequence[BondType] = (BondType.SINGLE, BondType.DOUBLE, BondType.TRIPLE),
                   initial_molecules: Optional[Sequence[MolGraph]] = None,
                   pop_size: int = 20,
                   metrics: Optional[List[str]] = None,
                   num_trials: int = 1,
                   trial_timeout: Optional[int] = None,
                   trial_iterations: Optional[int] = None,
                   visualize: bool = False,
                   save_history: bool = True,
                   pretrain_dir: Optional[str] = None,
                   ):
    metrics = metrics or ['qed_score']
    optimizer_id = optimizer_cls.__name__.lower()[:3]
    experiment_id = f'Experiment [optimizer={optimizer_id} metrics={", ".join(metrics)} pop_size={pop_size}]'
    exp_name = f'{optimizer_id}_{adaptive_kind.value}_popsize{pop_size}_min{trial_timeout}_{"_".join(metrics)}'

    atom_types = atom_types or ['C', 'N', 'O', 'F', 'P', 'S', 'Cl', 'Br']
    trial_results = []
    trial_histories = []
    trial_timedelta = timedelta(minutes=trial_timeout) if trial_timeout else None

    for trial in range(num_trials):
        optimizer, objective = optimizer_setup(optimizer_cls,
                                               adaptive_kind,
                                               max_heavy_atoms,
                                               atom_types,
                                               bond_types,
                                               trial_timedelta,
                                               trial_iterations,
                                               pop_size,
                                               metrics,
                                               initial_molecules)
        if pretrain_dir:
            pretrain_agent(optimizer, objective, pretrain_dir)

        found_graphs = optimizer.optimise(objective)
        history = optimizer.history

        if visualize:
            molecules = [MolAdapter().restore(graph) for graph in found_graphs]
            save_dir = Path('visualisations') / exp_name / f'trial_{trial}'
            visualize_results(set(molecules), objective, history, save_dir)
        if save_history:
            result_dir = Path('results') / exp_name
            result_dir.mkdir(parents=True, exist_ok=True)
            history.save(result_dir / f'history_trial_{trial}.json')
        trial_results.extend(history.final_choices)
        trial_histories.append(history)

    # Compute mean & std for metrics of trials
    ff = objective.format_fitness
    trial_metrics = np.array([ind.fitness.values for ind in trial_results])
    trial_metrics_mean = trial_metrics.mean(axis=0)
    trial_metrics_std = trial_metrics.std(axis=0)
    print(f'Experiment {experiment_id}\n'
          f'finished with metrics:\n'
          f'mean={ff(trial_metrics_mean)}\n'
          f' std={ff(trial_metrics_std)}')


def plot_experiment_comparison(experiment_ids: Sequence[str], metric_id: int = 0, results_dir='./results'):
    root = Path(results_dir)
    histories = {}
    for exp_name in experiment_ids:
        trials = []
        for history_filename in os.listdir(root / exp_name):
            if history_filename.startswith('history'):
                history = OptHistory.load(root / exp_name / history_filename)
                trials.append(history)
        histories[exp_name] = trials
        print(f'Loaded {len(trials)} trial histories for experiment: {exp_name}')
    # Visualize
    MultipleFitnessLines.from_histories(histories).visualize(metric_id=metric_id)
    return histories


if __name__ == '__main__':
    run_experiment(molecule_search_setup,
                   adaptive_kind=MutationAgentTypeEnum.bandit,
                   max_heavy_atoms=38,
                   trial_timeout=6,
                   pop_size=50,
                   visualize=True,
                   num_trials=5,
                   pretrain_dir=os.path.join(project_root(), 'examples', 'molecule_search', 'histories')
                   )



import json
import os
from typing import Optional, List

import joblib
import numpy as np
import pandas as pd
from guacamol.assess_goal_directed_generation import assess_goal_directed_generation
from guacamol.goal_directed_generator import GoalDirectedGenerator
from guacamol.scoring_function import ScoringFunction
from joblib import delayed
from rdkit.Chem import Draw, MolFromSmiles
from rdkit.Chem.rdchem import BondType

from examples.molecule_search.mol_adapter import MolAdapter
from examples.molecule_search.mol_advisor import MolChangeAdvisor
from examples.molecule_search.mol_graph import MolGraph
from examples.molecule_search.mol_graph_parameters import MolGraphRequirements
from examples.molecule_search.mol_mutations import CHEMICAL_MUTATIONS
from golem.core.optimisers.adaptive.operator_agent import MutationAgentTypeEnum
from golem.core.optimisers.genetic.gp_optimizer import EvoGraphOptimizer
from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters
from golem.core.optimisers.genetic.operators.crossover import CrossoverTypesEnum
from golem.core.optimisers.genetic.operators.elitism import ElitismTypesEnum
from golem.core.optimisers.genetic.operators.inheritance import GeneticSchemeTypesEnum
from golem.core.optimisers.objective import Objective
from golem.core.optimisers.optimizer import GraphGenerationParams


def load_init_population(scoring_function: ScoringFunction,
                         n_jobs: int = -1,
                         path=".\\data\\guacamol_v1_all.smiles",
                         number_of_molecules=100):
    """ Original code:
     https://github.com/BenevolentAI/guacamol_baselines/blob/master/graph_ga/goal_directed_generation.py"""
    with open(path, "r") as f:
        smiles_list = f.readlines()
    joblist = [delayed(scoring_function.score)(smile) for smile in smiles_list]
    scores = joblib.Parallel(n_jobs=n_jobs)(joblist)
    scored_smiles = list(zip(scores, smiles_list))
    scored_smiles = sorted(scored_smiles, key=lambda x: x[0], reverse=True)
    best_smiles = [smile for score, smile in scored_smiles][:number_of_molecules]
    init_pop = [MolGraph.from_smiles(smile) for smile in best_smiles]
    return init_pop


class GolemMoleculeGenerator(GoalDirectedGenerator):
    """ You need to download Guacamol all smiles dataset from https://figshare.com/projects/GuacaMol/56639"""
    def __init__(self,
                 requirements: Optional[MolGraphRequirements] = None,
                 graph_gen_params: Optional[GraphGenerationParams] = None,
                 gp_params: Optional[GPAlgorithmParameters] = None,
                 trial: int = 0):
        self.requirements = requirements or MolGraphRequirements(
            max_heavy_atoms=50,
            available_atom_types=['C', 'N', 'O', 'F', 'P', 'S', 'Cl', 'Br'],
            bond_types=(BondType.SINGLE, BondType.DOUBLE, BondType.TRIPLE),
            early_stopping_timeout=np.inf,
            early_stopping_iterations=700,
            keep_n_best=4,
            timeout=None,
            num_of_generations=3000,
            keep_history=True,
            n_jobs=-1,
            history_dir=None)

        self.graph_gen_params = graph_gen_params or GraphGenerationParams(
            adapter=MolAdapter(),
            advisor=MolChangeAdvisor())

        self.gp_params = gp_params or GPAlgorithmParameters(
            pop_size=50,
            max_pop_size=50,
            multi_objective=False,
            genetic_scheme_type=GeneticSchemeTypesEnum.steady_state,
            elitism_type=ElitismTypesEnum.replace_worst,
            mutation_types=CHEMICAL_MUTATIONS,
            crossover_types=[CrossoverTypesEnum.none],
            adaptive_mutation_type=MutationAgentTypeEnum.bandit)
        self.trial = trial
        self.task_num = 0

    def generate_optimized_molecules(self, scoring_function: ScoringFunction, number_molecules: int,
                                     starting_population: Optional[List[str]] = None) -> List[str]:
        objective = Objective(
            quality_metrics=lambda mol: -scoring_function.score(mol.get_smiles(aromatic=True)),
            is_multi_objective=False
        )
        self.gp_params.pop_size = max(number_molecules, 100)
        self.gp_params.max_pop_size = self.gp_params.pop_size

        initial_graphs = load_init_population(scoring_function, number_of_molecules=self.gp_params.pop_size)
        initial_graphs = self.graph_gen_params.adapter.adapt(initial_graphs)

        # Build the optimizer
        optimiser = EvoGraphOptimizer(objective,
                                      initial_graphs,
                                      self.requirements,
                                      self.graph_gen_params,
                                      self.gp_params)
        optimiser.optimise(objective)
        history = optimiser.history
        history.save(os.path.join(os.curdir, f'history_trial_{self.trial}_task{self.task_num}.json'))
        self.task_num += 1

        # Take only the first graph's appearance in history
        individuals \
            = list({hash(self.graph_gen_params.adapter.restore(ind.graph)): ind
                    for gen in history.generations
                    for ind in reversed(list(gen))}.values())

        top_individuals = sorted(individuals,
                                 key=lambda pos_ind: pos_ind.fitness, reverse=True)[:number_molecules]
        top_smiles = [MolAdapter().restore(ind.graph).get_smiles(aromatic=True) for ind in top_individuals]
        return top_smiles


def visualize(path: str):
    with open(path) as json_file:
        results = json.load(json_file)
    print(f"Guacamol version: {results['guacamol_version']} \n"
          f"Benchmark version: {results['benchmark_suite_version']} \n")
    results = results['results']
    for result in results:
        generated_molecules, scores = [[MolFromSmiles(smile) for smile, score in result['optimized_molecules'][:12]],
                                       [round(score, 3) for smile, score in result['optimized_molecules'][:12]]]
        benchmark_name = result['benchmark_name']
        scores = [f"{benchmark_name} : {score}" for score in scores]
        image = Draw.MolsToGridImage(generated_molecules,
                                     legends=scores,
                                     molsPerRow=min(4, len(generated_molecules)),
                                     subImgSize=(1000, 1000),
                                     legendFontSize=50)
        image.show()
        image.save(f'{benchmark_name}_results.png')


def get_launch_statistics(paths: List[str]):
    results = []
    for path in paths:
        with open(path) as json_file:
            results.append(json.load(json_file)['results'])

    column_names = ['benchmark', 'mean', 'std', 'min', 'max', 'mean_time']

    df = pd.DataFrame(columns=column_names)

    for bench_num in range(20):
        benchmark = results[0][bench_num]['benchmark_name']
        scores = []
        time_spent = []
        for result in results:
            scores.append(result[bench_num]['score'])
            time_spent.append(result[bench_num]['execution_time'])
        bench_result = pd.DataFrame(data=[[benchmark,
                                           np.mean(scores),
                                           np.std(scores),
                                           np.min(scores),
                                           np.max(scores),
                                           np.mean(time_spent)]],
                                    columns=column_names)
        df = pd.concat([df, bench_result], ignore_index=True, axis=0)
    pd.set_option('display.max_columns', None)
    print(df)
    df.to_csv('results.csv')


if __name__ == '__main__':
    # one launch takes more than 24h
    for launch in range(1):
        print(f'\nLaunch_num {launch}\n')
        assess_goal_directed_generation(GolemMoleculeGenerator(trial=launch),
                                        benchmark_version='v2',
                                        json_output_file=f'output_goal_directed_{launch}.json')
    visualize('output_goal_directed_1.json')
    get_launch_statistics([f'output_goal_directed_{launch}.json' for launch in range(4)])



import inspect
import os
import time

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from golem.core.dag.graph import Graph
from golem.core.optimisers.optimizer import GraphOptimizer
from golem.utilities.requirements_notificator import warn_requirement

try:
    import objgraph

    from memory_profiler import memory_usage
except ImportError:
    warn_requirement('objgraph', 'golem[profilers]', should_raise=True)


class MemoryProfiler:
    """Visual interpretation of memory usage. Create two ``png`` files.

    Args:
        function: function to profile.
        path: path to save profiling result.
        list args: arguments for function in array format.
        dict kwargs: arguments for function in dictionary format.
        list roots: array with classes used as ROOT nodes in the call-graph.
        int max_depth: maximum depth of graph.
    """

    def __init__(self, function, path: str, args=None, kwargs=None, roots=None, max_depth: int = 7,
                 visualization=False):
        if args is None:
            args = []

        if kwargs is None:
            kwargs = {}

        self.folder = os.path.abspath(path)

        if not os.path.exists(self.folder):
            os.mkdir(self.folder)

        if roots is None:
            roots = [Graph, GraphOptimizer]

        if visualization:
            # # Create the plot of the memory time dependence.
            # self._create_memory_plot(function, args, kwargs)

            # Create call graph.
            self._create_memory_graph(roots, max_depth)

    def _create_memory_plot(self, function, args, kwargs, interval: float = 0.1):
        start_time = time.time()
        mem_res = memory_usage((function, args, kwargs), interval=interval)
        total_time = time.time() - start_time

        length = len(mem_res)
        division = total_time / length
        time_split = np.linspace(division, total_time, length)

        max_index = np.argmax(mem_res)

        plt.figure(figsize=(12, 8))

        sns.set_style("whitegrid")

        ax = sns.lineplot((time_split, mem_res), linewidth=3, marker="X", markersize=8)
        ax.axhline(mem_res[max_index], ls='--', color='red')
        ax.axvline(time_split[max_index], ls='--', color='red')

        plt.legend(labels=[f'interval {interval}', 'max memory usage'])

        ax.set_title('Memory time dependence', size=25)

        ax.set_xlabel("time [s]", fontsize=16)
        ax.set_ylabel("memory used [KB]", fontsize=16)

        filename = os.path.join(self.folder, 'memory_plot.png')
        ax.figure.savefig(filename)

    def _create_memory_graph(self, roots, max_depth):
        filename = os.path.join(self.folder, 'memory_graph.png')

        objgraph.show_refs(
            roots,
            max_depth=max_depth,
            filename=filename,
            filter=lambda x: not type(x).__name__ in ['module', 'str', 'code', 'IPythonKernel', '_Helper',
                                                      'builtin_function_or_method', 'type', '_Printer', 'Printer',
                                                      'float64', 'float', 'int'],
            highlight=lambda x: inspect.isclass(x) or inspect.isfunction(x),
            refcounts=True
        )



from copy import deepcopy
from typing import TYPE_CHECKING

from golem.core.dag.graph_utils import ordered_subnodes_hierarchy
from golem.core.optimisers.genetic.operators.operator import PopulationT, EvaluationOperator, Operator
from golem.core.optimisers.graph import OptGraph, OptNode
from golem.core.optimisers.opt_history_objects.individual import Individual
from golem.core.optimisers.opt_history_objects.parent_operator import ParentOperator
from golem.core.optimisers.optimizer import GraphGenerationParams
from golem.utilities.data_structures import ComparableEnum as Enum

if TYPE_CHECKING:
    from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters


class RegularizationTypesEnum(Enum):
    none = 'none'
    decremental = 'decremental'


class Regularization(Operator):
    def __init__(self, parameters: 'GPAlgorithmParameters',
                 graph_generation_params: GraphGenerationParams):
        super().__init__(parameters=parameters)
        self.graph_generation_params = graph_generation_params

    def __call__(self, population: PopulationT, evaluator: EvaluationOperator) -> PopulationT:
        regularization_type = self.parameters.regularization_type
        if regularization_type is RegularizationTypesEnum.decremental:
            return self._decremental_regularization(population, evaluator)
        elif regularization_type is RegularizationTypesEnum.none:
            return population
        else:
            raise ValueError(f'Required regularization type not found: {regularization_type}')

    def _decremental_regularization(self, population: PopulationT, evaluator: EvaluationOperator) -> PopulationT:
        size = self.parameters.pop_size
        additional_inds = []
        prev_nodes_ids = set()
        for ind in population:
            prev_nodes_ids.add(ind.graph.descriptive_id)
            parent_operator = ParentOperator(type_='regularization',
                                             operators='decremental_regularization',
                                             parent_individuals=ind)
            subtree_inds = [Individual(OptGraph(deepcopy(ordered_subnodes_hierarchy(node))), parent_operator,
                                       metadata=self.requirements.static_individual_metadata)
                            for node in ind.graph.nodes
                            if Regularization._is_fitted_subtree(self.graph_generation_params.adapter.restore(node)) and
                            node.descriptive_id not in prev_nodes_ids]

            additional_inds.extend(subtree_inds)
            prev_nodes_ids.update(subtree.graph.root_node.descriptive_id for subtree in subtree_inds)

        additional_inds = [ind for ind in additional_inds if self.graph_generation_params.verifier(ind.graph)]
        evaluator(additional_inds)
        additional_inds.extend(population)
        if len(additional_inds) > size:
            additional_inds = sorted(additional_inds, key=lambda ind: ind.fitness, reverse=True)[:size]

        return additional_inds

    # TODO: remove this hack (e.g. provide smth like FitGraph with fit/unfit interface)
    @staticmethod
    def _is_fitted_subtree(node: OptNode) -> bool:
        return node.nodes_from and hasattr(node, 'fitted_operation') and node.fitted_operation



import dataclasses
import datetime
from dataclasses import dataclass, field
from numbers import Number
from typing import Optional

from golem.core.paths import default_data_dir
from golem.utilities.utilities import determine_n_jobs


@dataclass
class OptimizationParameters:
    """Defines general algorithm-independent parameters of the composition process
    (like stop condition, validation, timeout, logging etc.)

    Options related to stop condition:

    :param num_of_generations: maximum number of optimizer generations
    :param timeout: max time in minutes available for composition process
    :param early_stopping_iterations: for early stopping.

        Optional max number of stagnating
        iterations for early stopping. If both early_stopping options are None,
        then do not use early stopping.

    :param early_stopping_timeout: for early stopping.

        Optional duration (in minutes) of stagnating
        optimization for early stopping. If both early_stopping options are None,
        then do not use early stopping.

    Infrastructure options (logging, performance)

    :param keep_n_best: number of the best individuals of previous generation to keep in next generation
    :param max_graph_fit_time: time constraint for evaluation of each graph (datetime.timedelta)
    :param n_jobs: num of n_jobs
    :param show_progress: bool indicating whether to show progress using tqdm or not
    :param collect_intermediate_metric: save metrics for intermediate (non-root) nodes in graph
    :param parallelization_mode: identifies the way to parallelize population evaluation

    History options:

    :param keep_history: if True, then save generations to history; if False, don't keep history.
    :param history_dir: directory for saving optimization history, optional.

        If the path is relative, then save relative to `default_data_dir`.
        If absolute -- then save directly by specified path.
        If None -- do not save the history to disk and keep it only in-memory.
    """

    num_of_generations: Optional[int] = None
    timeout: Optional[datetime.timedelta] = datetime.timedelta(minutes=5)
    early_stopping_iterations: Optional[int] = 50
    early_stopping_timeout: Optional[float] = 5

    keep_n_best: int = 1
    max_graph_fit_time: Optional[datetime.timedelta] = None
    n_jobs: int = 1
    show_progress: bool = True
    collect_intermediate_metric: bool = False
    parallelization_mode: str = 'populational'
    static_individual_metadata: dict = field(default_factory=lambda: {
        'use_input_preprocessing': True
    })

    keep_history: bool = True
    history_dir: Optional[str] = field(default_factory=default_data_dir)
    agent_dir: Optional[str] = field(default_factory=default_data_dir)


@dataclass
class GraphRequirements(OptimizationParameters):
    """Defines restrictions and requirements on final graphs.

    Restrictions on final graphs:

    :param start_depth: start value of adaptive tree depth
    :param max_depth: max depth of the resulting graph
    :param min_arity: min number of parents for node
    :param max_arity: max number of parents for node
    """

    start_depth: int = 3
    max_depth: int = 10
    min_arity: int = 2
    max_arity: int = 4

    def __post_init__(self):
        # check and convert n_jobs to non-negative
        self.n_jobs = determine_n_jobs(self.n_jobs)

        for field_name, field_value in dataclasses.asdict(self).items():
            if isinstance(field_value, Number) and field_value < 0:
                raise ValueError(f'Value of {field_name} must be non-negative')



from random import choice
from typing import Optional

from examples.molecule_search.mol_graph import MolGraph
from examples.molecule_search.mol_graph_parameters import MolGraphRequirements
from examples.molecule_search.utils import get_free_electrons_num, get_functional_group
from golem.core.optimisers.optimization_parameters import GraphRequirements
from golem.core.optimisers.optimizer import GraphGenerationParams, AlgorithmParameters


def add_atom(mol_graph: MolGraph,
             requirements: MolGraphRequirements,
             graph_gen_params: GraphGenerationParams,
             parameters: Optional[AlgorithmParameters] = None):
    """ Adds new atom to a molecule. """
    atoms_to_connect = graph_gen_params.advisor.propose_connection_point(mol_graph)
    if atoms_to_connect and mol_graph.heavy_atoms_number < requirements.max_heavy_atoms:
        connect_to_atom_id = int(choice(atoms_to_connect))
        connect_to_atom = mol_graph.get_rw_molecule().GetAtomWithIdx(connect_to_atom_id)

        atom_types_to_add = graph_gen_params.advisor.propose_parent(connect_to_atom, requirements.available_atom_types)
        if atom_types_to_add:
            new_atom = choice(atom_types_to_add)
            mol_graph.add_atom(new_atom)

            new_atom_id = mol_graph.heavy_atoms_number - 1
            mol_graph.set_bond(new_atom_id, connect_to_atom_id)
            return mol_graph
    else:
        return mol_graph


def delete_atom(mol_graph: MolGraph,
                requirements: GraphRequirements,
                graph_gen_params: GraphGenerationParams,
                parameters: Optional[AlgorithmParameters] = None):
    """ Removes atom from a molecule """
    atoms_to_delete = graph_gen_params.advisor.propose_atom_removal(mol_graph)
    if atoms_to_delete:
        atom_to_delete_id = int(choice(atoms_to_delete))
        mol_graph.remove_atom(atom_to_delete_id)
    return mol_graph


def replace_atom(mol_graph: MolGraph,
                 requirements: MolGraphRequirements,
                 graph_gen_params: GraphGenerationParams,
                 parameters: Optional[AlgorithmParameters] = None):
    """ Replaces one atom with another """
    atom_to_replace = choice(mol_graph.get_rw_molecule().GetAtoms())
    possible_substitutions = graph_gen_params.advisor.propose_change(atom_to_replace, requirements.available_atom_types)
    if possible_substitutions:
        new_atom_type = choice(possible_substitutions)
        mol_graph.replace_atom(atom_to_replace.GetIdx(), new_atom_type)
    return mol_graph


def replace_bond(mol_graph: MolGraph,
                 requirements: MolGraphRequirements,
                 graph_gen_params: GraphGenerationParams,
                 parameters: Optional[AlgorithmParameters] = None):
    """ Changes degree of a bond """
    molecule = mol_graph.get_rw_molecule()
    atom_pairs_to_connect = graph_gen_params.advisor.propose_connection(mol_graph)

    if atom_pairs_to_connect:
        pair_to_connect = choice(atom_pairs_to_connect)
        from_atom = molecule.GetAtomWithIdx(pair_to_connect[0])
        to_atom = molecule.GetAtomWithIdx(pair_to_connect[1])

        pair_free_electrons = min(get_free_electrons_num(from_atom), get_free_electrons_num(to_atom))

        current_bond = molecule.GetBondBetweenAtoms(*pair_to_connect)
        current_bond_degree = current_bond.GetBondTypeAsDouble() if current_bond else 0

        max_bond_degree = current_bond_degree + pair_free_electrons
        possible_bonds = [bond for bond in requirements.bond_types if int(bond) <= max_bond_degree]
        if possible_bonds:
            bond_to_replace = choice(possible_bonds)
            mol_graph.set_bond(bond_type=bond_to_replace, *pair_to_connect)

    return mol_graph


def delete_bond(mol_graph: MolGraph,
                requirements: GraphRequirements,
                graph_gen_params: GraphGenerationParams,
                parameters: Optional[AlgorithmParameters] = None):
    """ Deletes a bond """
    atom_pairs_to_disconnect = graph_gen_params.advisor.propose_disconnection(mol_graph)
    if atom_pairs_to_disconnect:
        pair_to_disconnect = choice(atom_pairs_to_disconnect)
        mol_graph.remove_bond(*pair_to_disconnect)
    return mol_graph


def cut_atom(mol_graph: MolGraph,
             requirements: GraphRequirements,
             graph_gen_params: GraphGenerationParams,
             parameters: Optional[AlgorithmParameters] = None):
    """ Removes an atom if it has only two neighbors and the neighbors are disconnected.
    Neighbors are connected after the atom removal. """
    atoms_to_cut = graph_gen_params.advisor.propose_cut(mol_graph)
    molecule = mol_graph.get_rw_molecule()
    if atoms_to_cut:
        atom_to_cut = choice(atoms_to_cut)
        neighbors_id = [neighbor.GetIdx() for neighbor in molecule.GetAtomWithIdx(atom_to_cut).GetNeighbors()]
        mol_graph.set_bond(*neighbors_id, update_representation=False)
        mol_graph.remove_atom(atom_to_cut)
    return mol_graph


def insert_carbon(mol_graph: MolGraph,
                  requirements: MolGraphRequirements,
                  graph_gen_params: GraphGenerationParams,
                  parameters: Optional[AlgorithmParameters] = None):
    """ Inserts carbon between two connected atoms splitting the bond between them. """
    bonds_to_split = graph_gen_params.advisor.propose_bond_to_split(mol_graph)
    if bonds_to_split and mol_graph.heavy_atoms_number < requirements.max_heavy_atoms:
        bond_to_split = choice(bonds_to_split)
        mol_graph.remove_bond(*bond_to_split, update_representation=False)
        mol_graph.add_atom('C')
        carbon_id = mol_graph.heavy_atoms_number - 1
        mol_graph.set_bond(bond_to_split[0], carbon_id, update_representation=False)
        mol_graph.set_bond(bond_to_split[1], carbon_id)
    return mol_graph


def remove_group(mol_graph: MolGraph,
                 requirements: MolGraphRequirements,
                 graph_gen_params: GraphGenerationParams,
                 parameters: Optional[AlgorithmParameters] = None):
    """ Removes a functional group. """
    bridges_to_remove = graph_gen_params.advisor.propose_group(mol_graph)
    if bridges_to_remove:
        bridge = choice(bridges_to_remove)
        group = get_functional_group(mol_graph, bridge)
        for atom_id in sorted(group, reverse=True):
            mol_graph.remove_atom(atom_id, update_representation=False)
        mol_graph.update_representation()
    return mol_graph


def move_group(mol_graph: MolGraph,
               requirements: MolGraphRequirements,
               graph_gen_params: GraphGenerationParams,
               parameters: Optional[AlgorithmParameters] = None):
    """ Moves a functional group to another atom. """
    molecule = mol_graph.get_rw_molecule()
    bridges_to_move = graph_gen_params.advisor.propose_group(mol_graph)
    if bridges_to_move:
        bridge = choice(bridges_to_move)
        current_bond = molecule.GetBondBetweenAtoms(*bridge)
        group = get_functional_group(mol_graph, bridge)

        atoms_to_reconnect_to = graph_gen_params.advisor.propose_connection_point(mol_graph, current_bond)
        atoms_to_reconnect_to = [atom_idx for atom_idx in atoms_to_reconnect_to if atom_idx not in group.union(bridge)]
        if atoms_to_reconnect_to:
            new_endpoint = int(choice(atoms_to_reconnect_to))
            mol_graph.remove_bond(*bridge, update_representation=False)
            mol_graph.set_bond(new_endpoint, bridge[1], bond_type=current_bond.GetBondType())
    return mol_graph


CHEMICAL_MUTATIONS = [add_atom,
                      delete_atom,
                      replace_atom,
                      replace_bond,
                      delete_bond,
                      cut_atom,
                      insert_carbon,
                      remove_group,
                      move_group]



import functools
import os
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Union, Sequence, Tuple

import matplotlib as mpl
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.widgets import Button

from golem.core.log import default_log
from golem.core.optimisers.fitness import null_fitness
from golem.core.optimisers.opt_history_objects.individual import Individual
from golem.core.paths import default_data_dir
from golem.visualisation.opt_history.history_visualization import HistoryVisualization
from golem.visualisation.opt_history.utils import show_or_save_figure


def with_alternate_matplotlib_backend(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        default_mpl_backend = mpl.get_backend()
        try:
            mpl.use('TKAgg')
            return func(*args, **kwargs)
        except ImportError as e:
            default_log(prefix='Requirements').warning(e)
        finally:
            mpl.use(default_mpl_backend)

    return wrapper


def setup_fitness_plot(axis: plt.Axes, xlabel: str, title: Optional[str] = None, with_legend: bool = False):
    if axis is None:
        fig, axis = plt.subplots()

    if with_legend:
        axis.legend()
    axis.set_ylabel('Fitness')
    axis.set_xlabel(xlabel)
    axis.set_title(title)
    axis.grid(axis='y')


def plot_fitness_line_per_time(axis: plt.Axes, generations, label: Optional[str] = None,
                               with_generation_limits: bool = True) \
        -> Dict[int, Individual]:
    best_fitness = null_fitness()
    gen_start_times = []
    best_individuals = {}

    start_time = datetime.fromisoformat(
        min(generations[0], key=lambda ind: ind.metadata['evaluation_time_iso']).metadata[
            'evaluation_time_iso'])
    end_time_seconds = (datetime.fromisoformat(
        max(generations[-1], key=lambda ind: ind.metadata['evaluation_time_iso']).metadata[
            'evaluation_time_iso']) - start_time).seconds

    for gen_num, gen in enumerate(generations):
        gen_start_times.append(1e10)
        gen_sorted = sorted(gen, key=lambda ind: ind.metadata['evaluation_time_iso'])
        for ind in gen_sorted:
            if ind.native_generation != gen_num:
                continue
            evaluation_time = (datetime.fromisoformat(ind.metadata['evaluation_time_iso']) - start_time).seconds
            if evaluation_time < gen_start_times[gen_num]:
                gen_start_times[gen_num] = evaluation_time
            if ind.fitness > best_fitness:
                best_individuals[evaluation_time] = ind
                best_fitness = ind.fitness

    best_eval_times, best_fitnesses = np.transpose(
        [(evaluation_time, abs(individual.fitness.value))
         for evaluation_time, individual in best_individuals.items()])

    best_eval_times = list(best_eval_times)
    best_fitnesses = list(best_fitnesses)

    if best_eval_times[-1] != end_time_seconds:
        best_fitnesses.append(abs(best_fitness.value))
        best_eval_times.append(end_time_seconds)
    gen_start_times.append(end_time_seconds)

    axis.step(best_eval_times, best_fitnesses, where='post', label=label)

    if with_generation_limits:
        axis_gen = axis.twiny()
        axis_gen.set_xlim(axis.get_xlim())
        axis_gen.set_xticks(gen_start_times, list(range(len(gen_start_times) - 1)) + [''])
        axis_gen.locator_params(nbins=10)
        axis_gen.set_xlabel('Generation')

        gen_ticks = axis_gen.get_xticks()
        prev_time = gen_ticks[0]
        axis.axvline(prev_time, color='k', linestyle='--', alpha=0.3)
        for i, next_time in enumerate(gen_ticks[1:]):
            axis.axvline(next_time, color='k', linestyle='--', alpha=0.3)
            if i % 2 == 0:
                axis.axvspan(prev_time, next_time, color='k', alpha=0.05)
            prev_time = next_time

    return best_individuals


def plot_fitness_line_per_generations(axis: plt.Axes, generations, label: Optional[str] = None) \
        -> Dict[int, Individual]:
    best_fitnesses, best_generations, best_individuals = find_best_running_fitness(generations, metric_id=0)
    axis.step(best_generations, best_fitnesses, where='post', label=label)
    axis.set_xticks(range(len(generations)))
    axis.locator_params(nbins=10)
    return best_individuals


class FitnessLine(HistoryVisualization):
    def visualize(self, save_path: Optional[Union[os.PathLike, str]] = None, dpi: Optional[int] = None,
                  per_time: Optional[bool] = None):
        """ Visualizes the best fitness values during the evolution in the form of line.
        :param save_path: path to save the visualization. If set, then the image will be saved,
            and if not, it will be displayed.
        :param dpi: DPI of the output figure.
        :param per_time: defines whether to show time grid if it is available in history.
        """
        save_path = save_path or self.get_predefined_value('save_path')
        dpi = dpi or self.get_predefined_value('dpi')
        per_time = per_time if per_time is not None else self.get_predefined_value('per_time') or False

        fig, ax = plt.subplots(figsize=(6.4, 4.8), facecolor='w')
        if per_time:
            xlabel = 'Time, s'
            plot_fitness_line_per_time(ax, self.history.generations)
        else:
            xlabel = 'Generation'
            plot_fitness_line_per_generations(ax, self.history.generations)
        setup_fitness_plot(ax, xlabel)
        show_or_save_figure(fig, save_path, dpi)


class FitnessLineInteractive(HistoryVisualization):

    @with_alternate_matplotlib_backend
    def visualize(self, save_path: Optional[Union[os.PathLike, str]] = None, dpi: Optional[int] = None,
                  per_time: Optional[bool] = None, graph_show_kwargs: Optional[Dict[str, Any]] = None):
        """ Visualizes the best fitness values during the evolution in the form of line.
        Additionally, shows the structure of the best individuals and the moment of their discovering.
        :param save_path: path to save the visualization. If set, then the image will be saved, and if not,
            it will be displayed.
        :param dpi: DPI of the output figure.
        :param per_time: defines whether to show time grid if it is available in history.
        :param graph_show_kwargs: keyword arguments of `graph.show()` function.
        """

        save_path = save_path or self.get_predefined_value('save_path')
        dpi = dpi or self.get_predefined_value('dpi')
        per_time = per_time if per_time is not None else self.get_predefined_value('per_time') or False
        graph_show_kwargs = graph_show_kwargs or self.get_predefined_value('graph_show_params') or {}

        graph_show_kwargs = graph_show_kwargs or self.visuals_params.get('graph_show_params') or {}

        fig, axes = plt.subplots(1, 2, figsize=(15, 10))
        ax_fitness, ax_graph = axes

        if per_time:
            x_label = 'Time, s'
            x_template = 'time {} s'
            plot_func = plot_fitness_line_per_time
        else:
            x_label = 'Generation'
            x_template = 'generation {}'
            plot_func = plot_fitness_line_per_generations

        best_individuals = plot_func(ax_fitness, self.history.generations)
        setup_fitness_plot(ax_fitness, x_label)

        ax_graph.axis('off')

        class InteractivePlot:
            temp_path = Path(default_data_dir(), 'current_graph.png')

            def __init__(self, best_individuals: Dict[int, Individual]):
                self.best_x: List[int] = list(best_individuals.keys())
                self.best_individuals: List[Individual] = list(best_individuals.values())
                self.index: int = len(self.best_individuals) - 1
                self.time_line = ax_fitness.axvline(self.best_x[self.index], color='r', alpha=0.7)
                self.graph_images: List[np.ndarray] = []
                self.generate_graph_images()
                self.update_graph()

            def generate_graph_images(self):
                for ind in self.best_individuals:
                    graph = ind.graph
                    graph.show(self.temp_path, **graph_show_kwargs)
                    self.graph_images.append(plt.imread(str(self.temp_path)))
                self.temp_path.unlink()

            def update_graph(self):
                ax_graph.imshow(self.graph_images[self.index])
                x = self.best_x[self.index]
                fitness = self.best_individuals[self.index].fitness
                ax_graph.set_title(f'The best graph at {x_template.format(x)}, fitness={fitness}')

            def update_time_line(self):
                self.time_line.set_xdata(self.best_x[self.index])

            def step_index(self, step: int):
                self.index = (self.index + step) % len(self.best_individuals)
                self.update_graph()
                self.update_time_line()
                plt.draw()

            def next(self, event):
                self.step_index(1)

            def prev(self, event):
                self.step_index(-1)

        callback = InteractivePlot(best_individuals)

        if not save_path:  # display buttons only for an interactive plot
            ax_prev = plt.axes([0.7, 0.05, 0.1, 0.075])
            ax_next = plt.axes([0.81, 0.05, 0.1, 0.075])
            b_next = Button(ax_next, 'Next')
            b_next.on_clicked(callback.next)
            b_prev = Button(ax_prev, 'Previous')
            b_prev.on_clicked(callback.prev)

        show_or_save_figure(fig, save_path, dpi)


def find_best_running_fitness(generations: Sequence[Sequence[Individual]],
                              metric_id: int = 0,
                              ) -> Tuple[List[float], List[int], Dict[int, Individual]]:
    """For each trial history per each generation find the best fitness *seen so far*.
    Returns tuple:
    - list of best seen metric up to that generation,
    - list of indices where current best individual belongs.
    - dict mapping of best index to best individuals
    """
    best_metric = np.inf  # Assuming metric minimization
    best_individuals = {}

    # Core logic
    for gen_num, gen in enumerate(generations):
        for ind in gen:
            if ind.native_generation != gen_num:
                continue
            target_metric = ind.fitness.values[metric_id]
            if target_metric <= best_metric:
                best_individuals[gen_num] = ind
                best_metric = target_metric

    # Additional unwrapping of the data for simpler plotting
    best_generations, best_metrics = np.transpose(
        [(gen_num, abs(individual.fitness.values[metric_id]))
         for gen_num, individual in best_individuals.items()])
    best_generations = list(best_generations)
    best_metrics = list(best_metrics)
    if best_generations[-1] != len(generations) - 1:
        best_metrics.append(abs(best_metric))
        best_generations.append(len(generations) - 1)

    return best_metrics, best_generations, best_individuals



from __future__ import annotations

from abc import abstractmethod
from copy import deepcopy
from typing import TYPE_CHECKING, TypeVar, Generic, Type, Optional, Dict, Any, Callable, Tuple, Sequence, Union

from golem.core.dag.graph import Graph
from golem.core.log import default_log
from golem.core.optimisers.graph import OptGraph, OptNode
from golem.core.adapter.adapt_registry import AdaptRegistry
from golem.core.optimisers.opt_history_objects.individual import Individual

if TYPE_CHECKING:
    from golem.core.optimisers.genetic.operators.operator import PopulationT

DomainStructureType = TypeVar('DomainStructureType')


class BaseOptimizationAdapter(Generic[DomainStructureType]):
    def __init__(self, base_graph_class: Type[DomainStructureType] = Graph):
        self._log = default_log(self)
        self.domain_graph_class = base_graph_class
        self.opt_graph_class = OptGraph

    def restore_func(self, fun: Callable) -> Callable:
        """Wraps native function so that it could accept domain graphs as arguments.

        Behavior: ``restore( f(Graph)->Graph ) => f'(DomainGraph)->DomainGraph``

        Implementation details.
        The method wraps callable into a function that transforms its args & return value.
        Arguments are transformed by ``adapt`` (that maps domain graphs to internal graphs).
        Return value is transformed by ``restore`` (that maps internal graphs to domain graphs).

        Args:
            fun: native function that accepts native args (i.e. optimization graph)

        Returns:
            Callable: domain function that can accept domain graphs
        """
        return _transform(fun, f_args=self.adapt, f_ret=self.restore)

    def adapt_func(self, fun: Callable) -> Callable:
        """Wraps domain function so that it could accept native optimization graphs
        as arguments. If the function was registered as native, it is returned as-is.
        ``AdaptRegistry`` is responsible for function registration.

        Behavior: ``adapt( f(DomainGraph)->DomainGraph ) => f'(Graph)->Graph``

        Implementation details.
        The method wraps callable into a function that transforms its args & return value.
        Arguments are transformed by ``restore`` (that maps internal graphs to domain graphs).
        Return value is transformed by ``adapt`` (that maps domain graphs to internal graphs).

        Args:
            fun: domain function that accepts domain graphs

        Returns:
            Callable: native function that can accept opt graphs
            and be used inside Optimizer
        """
        if AdaptRegistry.is_native(fun):
            return fun
        return _transform(fun, f_args=self.restore, f_ret=self.adapt)

    def adapt(self, item: Union[DomainStructureType, Sequence[DomainStructureType]]) \
            -> Union[Graph, Sequence[Graph]]:
        """Maps domain graphs to internal graph representation used by optimizer.
        Performs mapping only if argument has a type of domain graph.

        Args:
            item: a domain graph or sequence of them

        Returns:
            Graph | Sequence: mapped internal graph or sequence of them
        """
        if type(item) is self.domain_graph_class:
            return self._adapt(item)
        elif isinstance(item, Sequence) and type(item[0]) is self.domain_graph_class:
            return [self._adapt(graph) for graph in item]
        else:
            return item

    def restore(self, item: Union[Graph, Individual, PopulationT, Sequence[Graph]]) \
            -> Union[DomainStructureType, Sequence[DomainStructureType]]:
        """Maps graphs from internal representation to domain graphs.
        Performs mapping only if argument has a type of internal representation.

        Args:
            item: an internal graph representation or sequence of them

        Returns:
            Graph | Sequence: mapped domain graph or sequence of them
        """
        if type(item) is self.opt_graph_class:
            return self._restore(item)
        elif isinstance(item, Individual):
            return self._restore(item.graph, item.metadata)
        elif isinstance(item, Sequence) and isinstance(item[0], Individual):
            return [self._restore(ind.graph, ind.metadata) for ind in item]
        elif isinstance(item, Sequence) and isinstance(item[0], self.opt_graph_class):
            return [self._restore(graph) for graph in item]
        else:
            return item

    @abstractmethod
    def _adapt(self, adaptee: DomainStructureType) -> Graph:
        """Implementation of ``adapt`` for single graph."""
        raise NotImplementedError()

    @abstractmethod
    def _restore(self, opt_graph: Graph, metadata: Optional[Dict[str, Any]] = None) -> DomainStructureType:
        """Implementation of ``restore`` for single graph."""
        raise NotImplementedError()


class IdentityAdapter(BaseOptimizationAdapter[DomainStructureType]):
    """Identity adapter that performs no transformation, returning same graphs."""

    def _adapt(self, adaptee: DomainStructureType) -> Graph:
        return adaptee

    def _restore(self, opt_graph: Graph, metadata: Optional[Dict[str, Any]] = None) -> DomainStructureType:
        return opt_graph


class DirectAdapter(BaseOptimizationAdapter[DomainStructureType]):
    """Naive optimization adapter for arbitrary class that just overwrites __class__."""

    def __init__(self,
                 base_graph_class: Type[DomainStructureType] = OptGraph,
                 base_node_class: Type = OptNode):
        super().__init__(base_graph_class)
        self.domain_node_class = base_node_class

    def _adapt(self, adaptee: DomainStructureType) -> Graph:
        opt_graph = deepcopy(adaptee)
        opt_graph.__class__ = self.opt_graph_class
        for node in opt_graph.nodes:
            node.__class__ = OptNode
        return opt_graph

    def _restore(self, opt_graph: Graph, metadata: Optional[Dict[str, Any]] = None) -> DomainStructureType:
        obj = deepcopy(opt_graph)
        obj.__class__ = self.domain_graph_class
        for node in obj.nodes:
            node.__class__ = self.domain_node_class
        return obj


def _transform(fun: Callable, f_args: Callable, f_ret: Callable) -> Callable:
    """Wraps function by transforming its arguments and return value:
     ``f_args`` is called on each of the function arguments,
     ``f_ret`` is called on the return value of original function.

     This is a helper function used for adaption of callables by
     :py:class:`golem.core.adapter.adapter.BaseOptimizationAdapter`.

    Args:
        fun: function to be transformed
        f_args: argument transformation function
        f_ret: return value transformation function

    Returns:
        Callable: wrapped transformed function
    """

    if not isinstance(fun, Callable):
        raise ValueError(f'Expected Callable, got {type(fun)}')

    def adapted_fun(*args, **kwargs):
        adapted_args = (f_args(arg) for arg in args)
        adapted_kwargs = dict((kw, f_args(arg)) for kw, arg in kwargs.items())

        result = fun(*adapted_args, **adapted_kwargs)

        if result is None:
            adapted_result = None
        elif isinstance(result, Tuple):
            # In case when function returns not only Graph
            adapted_result = tuple(f_ret(result_item) for result_item in result)
        else:
            adapted_result = f_ret(result)
        return adapted_result

    return adapted_fun



from typing import Union, Hashable, Tuple, Sequence

from golem.core.dag.graph import Graph
from golem.core.optimisers.opt_history_objects.individual import Individual

ObsType = Union[Individual, Graph]
ActType = Hashable
# Trajectory step includes (past observation, action, reward)
TrajectoryStep = Tuple[Individual, ActType, float]
# Trajectory is a sequence of applied mutations and received rewards
GraphTrajectory = Sequence[TrajectoryStep]



from copy import deepcopy
from typing import Optional, Dict, Any

import networkx as nx

from examples.molecule_search.mol_graph import MolGraph
from examples.synthetic_graph_evolution.generators import nx_to_directed
from golem.core.adapter import BaseOptimizationAdapter
from golem.core.adapter.nx_adapter import BaseNetworkxAdapter
from golem.core.optimisers.graph import OptGraph


class MolAdapter(BaseOptimizationAdapter):
    def __init__(self):
        super().__init__(base_graph_class=MolGraph)
        self.nx_adapter = BaseNetworkxAdapter()

    def _restore(self, opt_graph: OptGraph, metadata: Optional[Dict[str, Any]] = None) -> MolGraph:
        digraph = self.nx_adapter.restore(opt_graph)
        # to ensure backward compatibility with old individuals without 'nxid' field in nodes
        if not any(x is None for x in list(dict(digraph.nodes(data='nxid')).values())):
            # return to previous node indexing
            digraph = nx.relabel_nodes(digraph, dict(digraph.nodes(data='nxid')))
        digraph = restore_edges_params_from_nodes(digraph)
        nx_graph = digraph.to_undirected()
        mol_graph = MolGraph.from_nx_graph(nx_graph)
        return mol_graph

    def _adapt(self, adaptee: MolGraph) -> OptGraph:
        nx_graph = adaptee.get_nx_graph()
        digraph = nx_to_directed(nx_graph)
        digraph = store_edges_params_in_nodes(digraph)
        opt_graph = self.nx_adapter.adapt(digraph)
        return opt_graph


def store_edges_params_in_nodes(graph: nx.DiGraph) -> nx.DiGraph:
    graph = deepcopy(graph)
    edges_params = {}
    for node in graph.nodes():
        edge_params = {}
        for predecessor in graph.predecessors(node):
            edge_params.update({str(predecessor): graph.get_edge_data(predecessor, node)})
        edges_params.update({node: edge_params})
    nx.set_node_attributes(graph, edges_params, name='edges_params')
    return graph


def restore_edges_params_from_nodes(graph: nx.DiGraph) -> nx.DiGraph:
    graph = deepcopy(graph)
    edge_params_by_node = nx.get_node_attributes(graph, 'edges_params')
    all_edges_params = {}
    for node in graph.nodes():
        for predecessor in graph.predecessors(node):
            node_params = edge_params_by_node[node]
            # sometimes by unknown reason some nodes are encoded as int, some as str.
            # maybe that's deserialization messing up somewhere.
            edge_params = node_params.get(predecessor) or node_params.get(str(predecessor))
            if edge_params:
                all_edges_params[(predecessor, node)] = edge_params
    nx.set_edge_attributes(graph, all_edges_params)
    return graph



from abc import abstractmethod
from copy import deepcopy
from datetime import timedelta
from typing import TypeVar, Generic, Optional, Union, Sequence

import numpy as np

from golem.core.adapter import BaseOptimizationAdapter
from golem.core.adapter.adapter import IdentityAdapter
from golem.core.constants import MAX_TUNING_METRIC_VALUE
from golem.core.dag.graph_utils import graph_structure
from golem.core.log import default_log
from golem.core.optimisers.fitness import SingleObjFitness, MultiObjFitness
from golem.core.optimisers.graph import OptGraph
from golem.core.optimisers.objective import ObjectiveEvaluate, ObjectiveFunction
from golem.core.tuning.search_space import SearchSpace, convert_parameters
from golem.utilities.data_structures import ensure_wrapped_in_sequence

DomainGraphForTune = TypeVar('DomainGraphForTune')


class BaseTuner(Generic[DomainGraphForTune]):
    """
    Base class for hyperparameters optimization

    Args:
      objective_evaluate: objective to optimize
      adapter: the function for processing of external object that should be optimized
      search_space: SearchSpace instance
      iterations: max number of iterations
      early_stopping_rounds: Optional max number of stagnating iterations for early stopping.
      timeout: max time for tuning
      n_jobs: num of ``n_jobs`` for parallelization (``-1`` for use all cpu's)
      deviation: required improvement (in percent) of a metric to return tuned graph.
        By default, ``deviation=0.05``, which means that tuned graph will be returned
        if it's metric will be at least 0.05% better than the initial.
    """

    def __init__(self, objective_evaluate: ObjectiveFunction,
                 search_space: SearchSpace,
                 adapter: Optional[BaseOptimizationAdapter] = None,
                 iterations: int = 100,
                 early_stopping_rounds: Optional[int] = None,
                 timeout: timedelta = timedelta(minutes=5),
                 n_jobs: int = -1,
                 deviation: float = 0.05):
        self.iterations = iterations
        self.adapter = adapter or IdentityAdapter()
        self.search_space = search_space
        self.n_jobs = n_jobs
        if isinstance(objective_evaluate, ObjectiveEvaluate):
            objective_evaluate.eval_n_jobs = self.n_jobs
        self.objective_evaluate = self.adapter.adapt_func(objective_evaluate)
        self.deviation = deviation

        self.timeout = timeout
        self.early_stopping_rounds = early_stopping_rounds

        self._default_metric_value = MAX_TUNING_METRIC_VALUE
        self.was_tuned = False
        self.init_graph = None
        self.init_metric = None
        self.obtained_metric = None
        self.log = default_log(self)

    @abstractmethod
    def tune(self, graph: DomainGraphForTune) -> Union[DomainGraphForTune, Sequence[DomainGraphForTune]]:
        """
        Function for hyperparameters tuning on the graph

        Args:
          graph: domain graph for which hyperparameters tuning is needed

        Returns:
          Graph with optimized hyperparameters
          or pareto front of optimized graphs in case of multi-objective optimization
        """
        raise NotImplementedError()

    def init_check(self, graph: OptGraph) -> None:
        """
        Method get metric on validation set before start optimization

        Args:
          graph: graph to calculate objective
        """
        self.log.info('Hyperparameters optimization start: estimation of metric for initial graph')

        # Train graph
        self.init_graph = deepcopy(graph)

        self.init_metric = self.get_metric_value(graph=self.init_graph)
        self.log.message(f'Initial graph: {graph_structure(self.init_graph)} \n'
                         f'Initial metric: '
                         f'{list(map(lambda x: round(abs(x), 3), ensure_wrapped_in_sequence(self.init_metric)))}')

    def final_check(self, tuned_graphs: Union[OptGraph, Sequence[OptGraph]], multi_obj: bool = False) \
            -> Union[OptGraph, Sequence[OptGraph]]:
        """
        Method propose final quality check after optimization process

        Args:
          tuned_graphs: Tuned graph to calculate objective
          multi_obj: If optimization was multi objective.
        """
        self.log.info('Hyperparameters optimization finished')

        if multi_obj:
            return self._multi_obj_final_check(tuned_graphs)
        else:
            return self._single_obj_final_check(tuned_graphs)

    def _single_obj_final_check(self, tuned_graph: OptGraph):
        self.obtained_metric = self.get_metric_value(graph=tuned_graph)

        prefix_tuned_phrase = 'Return tuned graph due to the fact that obtained metric'
        prefix_init_phrase = 'Return init graph due to the fact that obtained metric'

        if np.isclose(self.obtained_metric, self._default_metric_value):
            self.obtained_metric = None

        # 0.05% deviation is acceptable
        deviation_value = (self.init_metric / 100.0) * self.deviation
        init_metric = self.init_metric + deviation_value * (-np.sign(self.init_metric))
        if self.obtained_metric is None:
            self.log.info(f'{prefix_init_phrase} is None. Initial metric is {abs(init_metric):.3f}')
            final_graph = self.init_graph
            final_metric = self.init_metric
        elif self.obtained_metric <= init_metric:
            self.log.info(f'{prefix_tuned_phrase} {abs(self.obtained_metric):.3f} equal or '
                          f'better than initial (+ {self.deviation}% deviation) {abs(init_metric):.3f}')
            final_graph = tuned_graph
            final_metric = self.obtained_metric
        else:
            self.log.info(f'{prefix_init_phrase} {abs(self.obtained_metric):.3f} '
                          f'worse than initial (+ {self.deviation}% deviation) {abs(init_metric):.3f}')
            final_graph = self.init_graph
            final_metric = self.init_metric
            self.obtained_metric = final_metric
        self.log.message(f'Final graph: {graph_structure(final_graph)}')
        if final_metric is not None:
            self.log.message(f'Final metric: {abs(final_metric):.3f}')
        else:
            self.log.message('Final metric is None')
        return final_graph

    def _multi_obj_final_check(self, tuned_graphs: Sequence[OptGraph]) -> Sequence[OptGraph]:
        self.obtained_metric = []
        final_graphs = []
        for tuned_graph in tuned_graphs:
            obtained_metric = self.get_metric_value(graph=tuned_graph)
            for e, value in enumerate(obtained_metric):
                if np.isclose(value, self._default_metric_value):
                    obtained_metric[e] = None
            if not MultiObjFitness(self.init_metric).dominates(MultiObjFitness(obtained_metric)):
                self.obtained_metric.append(obtained_metric)
                final_graphs.append(tuned_graph)
        if final_graphs:
            metrics_formatted = [str([round(x, 3) for x in metrics]) for metrics in self.obtained_metric]
            metrics_formatted = '\n'.join(metrics_formatted)
            self.log.message('Return tuned graphs with obtained metrics \n'
                             f'{metrics_formatted}')
        else:
            self.log.message('Initial metric dominates all found solutions. Return initial graph.')
            final_graphs = self.init_graph
        return final_graphs

    def get_metric_value(self, graph: OptGraph) -> Union[float, Sequence[float]]:
        """
        Method calculates metric for algorithm validation

        Args:
          graph: Graph to evaluate

        Returns:
          value of loss function
        """
        graph_fitness = self.objective_evaluate(graph)

        if isinstance(graph_fitness, SingleObjFitness):
            metric_value = graph_fitness.value
            if not graph_fitness.valid:
                return self._default_metric_value
            return metric_value

        elif isinstance(graph_fitness, MultiObjFitness):
            metric_values = graph_fitness.values
            for e, value in enumerate(metric_values):
                if value is None:
                    metric_values[e] = self._default_metric_value
            return metric_values

    @staticmethod
    def set_arg_graph(graph: OptGraph, parameters: dict) -> OptGraph:
        """ Method for parameters setting to a graph

        Args:
            graph: graph to which parameters should be assigned
            parameters: dictionary with parameters to set

        Returns:
            graph: graph with new hyperparameters in each node
        """
        # Set hyperparameters for every node
        for node_id, node in enumerate(graph.nodes):
            node_params = {key: value for key, value in parameters.items()
                           if key.startswith(f'{str(node_id)} || {node.name}')}

            if node_params is not None:
                BaseTuner.set_arg_node(graph, node_id, node_params)

        return graph

    @staticmethod
    def set_arg_node(graph: OptGraph, node_id: int, node_params: dict) -> OptGraph:
        """ Method for parameters setting to a graph

        Args:
            graph: graph which contains the node
            node_id: id of the node to which parameters should be assigned
            node_params: dictionary with labeled parameters to set

        Returns:
            graph with new hyperparameters in each node
        """

        # Remove label prefixes
        node_params = convert_parameters(node_params)

        # Update parameters in nodes
        graph.nodes[node_id].parameters = node_params

        return graph

    def _stop_tuning_with_message(self, message: str):
        self.log.message(message)
        self.obtained_metric = self.init_metric



from dataclasses import dataclass, field
from typing import List, Sequence

from rdkit.Chem.rdchem import BondType

from golem.core.optimisers.optimization_parameters import GraphRequirements


@dataclass
class MolGraphRequirements(GraphRequirements):
    max_heavy_atoms: int = 50
    available_atom_types: List[str] = field(default_factory=lambda: ['C', 'N', 'O', 'F', 'P', 'S', 'Cl', 'Br'])
    bond_types: Sequence[BondType] = field(default_factory=lambda: [BondType.SINGLE, BondType.DOUBLE, BondType.TRIPLE])



from typing import Any, Dict, Type
from uuid import UUID


def uuid_to_json(obj: UUID) -> Dict[str, Any]:
    return { 'hex': obj.hex }


def uuid_from_json(cls: Type[UUID], json_obj: Dict[str, Any]) -> UUID:
    return cls(json_obj['hex'])



from functools import partial
from typing import Callable, Optional, Sequence, Union, Iterable

from golem.core.optimisers.optimization_parameters import GraphRequirements
from golem.core.constants import MAX_GRAPH_GEN_ATTEMPTS
from golem.core.dag.graph import Graph
from golem.core.log import default_log
from golem.core.optimisers.graph import OptGraph
from golem.core.optimisers.optimizer import GraphGenerationParams

GenerationFunction = Callable[[], Graph]
InitialGraphsGenerator = Callable[[], Sequence[OptGraph]]


class InitialPopulationGenerator(InitialGraphsGenerator):
    """Generates initial population using three approaches.
    One is with initial graphs.
    Another is with initial graphs generation function which generates a graph
    that will be added to initial population.
    The third way is random graphs generation according to GraphGenerationParameters and OptimizationParameters.
    The last approach is applied when neither initial graphs nor initial graphs generation function were provided."""

    def __init__(self,
                 population_size: int,
                 generation_params: GraphGenerationParams,
                 requirements: GraphRequirements):
        self.pop_size = population_size
        self.requirements = requirements
        self.graph_generation_params = generation_params
        self.generation_function: Optional[GenerationFunction] = None
        self.initial_graphs: Optional[Sequence[Graph]] = None
        self.log = default_log(self)

    def __call__(self) -> Sequence[OptGraph]:
        pop_size = self.pop_size
        adapter = self.graph_generation_params.adapter

        if self.initial_graphs:
            if len(self.initial_graphs) > pop_size:
                self.initial_graphs = self.initial_graphs[:pop_size]
            return adapter.adapt(self.initial_graphs)

        if not self.generation_function:
            self.generation_function = partial(self.graph_generation_params.random_graph_factory, self.requirements)

        population = []
        for iter_num in range(MAX_GRAPH_GEN_ATTEMPTS):
            if len(population) == pop_size:
                break
            new_graph = self.generation_function()
            if new_graph not in population and self.graph_generation_params.verifier(new_graph):
                population.append(new_graph)
        else:
            self.log.warning(f'Exceeded max number of attempts for generating initial graphs, stopping.'
                             f'Generated {len(population)} instead of {pop_size} graphs.')
        return population

    def with_initial_graphs(self, initial_graphs: Union[Graph, Sequence[Graph]]):
        """Use initial graphs as initial population."""
        if isinstance(initial_graphs, Graph):
            self.initial_graphs = [initial_graphs]
        elif isinstance(initial_graphs, Iterable):
            self.initial_graphs = list(initial_graphs)
        else:
            raise ValueError(f'Incorrect type of initial_assumption: '
                             f'Sequence[Graph] or Graph needed, but has {type(initial_graphs)}')
        return self

    def with_custom_generation_function(self, generation_func: GenerationFunction):
        """Use custom graph generation function to create initial population."""
        self.generation_function = generation_func
        return self



from copy import copy
from functools import partial
from typing import Callable

from golem.utilities.singleton_meta import SingletonMeta


class AdaptRegistry(metaclass=SingletonMeta):
    """Registry of callables that require adaptation of argument/return values.
    AdaptRegistry together with :py:class:`golem.core.adapter.adapter.BaseOptimizationAdapter`
    enables automatic transformation between internal and domain graph representations.

    **Short description of the use-case**

    Operators & verification rules that operate on internal representation
    of graphs must be marked as native with decorator
    :py:func:`golem.core.adapter.adapt_registry.register_native`.

    Usually this is the case when users of the framework provide custom
    operators for internal optimization graphs. When custom operators
    operate on domain graphs, nothing is required.

    **Extended description**

    Optimiser operates with generic graph representation.
    Because of this any domain function requires adaptation
    of its graph arguments. Adapter can automatically adapt
    arguments to generic form in such cases.

    Important notions:

    * 'Domain' functions operate with domain-specific graphs.
    * 'Native' functions operate with generic graphs used by optimiser.
    * 'External' functions are functions defined by users of optimiser.

        Most notably, custom mutations and custom verifier rules.

    * 'Internal' functions are those defined by graph optimiser.

        Most notably, the default set of mutations and verifier rules.
        All internal functions are native.

    Adaptation registry usage and behavior:

    * Domain functions are adapted by default.
    * Native functions don't require adaptation of their arguments.
    * External functions are considered 'domain' functions by default.

        Hence, their arguments are adapted, unless users of optimiser
        exclude them from the process of automatic adaptation.
        It can be done by registering them as 'native'.

    AdaptRegistry can be safely used with multiprocessing
    insofar as all relevant functions are registered as native
    in the main process before child processes are started.
    """

    _native_flag_attr_name_ = '_adapter_is_optimizer_native'

    def __init__(self):
        self._registered_native_callables = []

    def register_native(self, fun: Callable) -> Callable:
        """Registers callable object as an internal function
        that can work with internal graph representation.
        Hence, it doesn't require adaptation when called by the optimiser.

        Implementation details.
        Works by setting a special attribute on the object.
        This attribute then is checked by ``is_native`` used by adapters.

        Args:
            fun: function or callable to be registered as native

        Returns:
            Callable: same function with special private attribute set
        """
        original_function = AdaptRegistry._get_underlying_func(fun)
        setattr(original_function, AdaptRegistry._native_flag_attr_name_, True)
        self._registered_native_callables.append(original_function)
        return fun

    def unregister_native(self, fun: Callable) -> Callable:
        """Unregisters callable object. See ``register_native``.

        Args:
            fun: function or callable to be unregistered as native

        Returns:
            Callable: same function with special private attribute unset
        """
        original_function = AdaptRegistry._get_underlying_func(fun)
        if hasattr(original_function, AdaptRegistry._native_flag_attr_name_):
            delattr(original_function, AdaptRegistry._native_flag_attr_name_)
            self._registered_native_callables.remove(original_function)
        return fun

    @staticmethod
    def is_native(fun: Callable) -> bool:
        """Tests callable object for a presence of specific attribute
        that tells that this function must not be restored with Adapter.

        Args:
            fun: tested Callable (function, method, functools.partial, or any callable object)

        Returns:
            bool: True if the callable was registered as native, False otherwise.
        """
        original_function = AdaptRegistry._get_underlying_func(fun)
        is_native = getattr(original_function, AdaptRegistry._native_flag_attr_name_, False)
        return is_native

    def clear_registered_callables(self):
        # copy is to avoid removing elements from list while iterating
        for f in copy(self._registered_native_callables):
            self.unregister_native(f)

    @staticmethod
    def _get_underlying_func(obj: Callable) -> Callable:
        """Recursively unpacks 'partial' and 'method' objects to get underlying function.

        Args:
            obj: callable to try unpacking

        Returns:
            Callable: unpacked function that underlies the callable, or the unchanged object itself
        """
        while True:
            if isinstance(obj, partial):  # if it is a 'partial'
                obj = obj.func
            elif hasattr(obj, '__func__'):  # if it is a 'method'
                obj = obj.__func__
            else:
                return obj  # return the unpacked underlying function or the original object


def register_native(fun: Callable) -> Callable:
    """Out-of-class version of the ``register_native``
    function that's intended to be used as a decorator.

    Args:
        fun: function or callable to be registered as native

    Returns:
        Callable: same function with special private attribute set
    """
    return AdaptRegistry().register_native(fun)



from typing import Sequence, Optional, Callable

from golem.core.adapter import BaseOptimizationAdapter
from golem.core.adapter.adapter import IdentityAdapter
from golem.core.dag.graph import Graph
from golem.core.log import default_log

# Validation rule can either return False or raise a ValueError to signal a failed check
VerifierRuleType = Callable[..., bool]


class VerificationError(ValueError):
    pass


class GraphVerifier:
    def __init__(self, rules: Sequence[VerifierRuleType] = (),
                 adapter: Optional[BaseOptimizationAdapter] = None,
                 raise_on_failure: bool = False):
        self._adapter = adapter or IdentityAdapter()
        self._rules = rules
        self._log = default_log(self)
        self._raise = raise_on_failure

    def __call__(self, graph: Graph) -> bool:
        return self.verify(graph)

    def verify(self, graph: Graph) -> bool:
        # Check if all rules pass
        adapt = self._adapter.adapt_func
        for rule in self._rules:
            try:
                if adapt(rule)(graph) is False:
                    return False
            except ValueError as err:
                msg = f'Graph verification failed with error <{err}> '\
                      f'for rule={rule} on graph={graph.descriptive_id}.'
                if self._raise:
                    raise VerificationError(msg)
                else:
                    self._log.debug(msg)
                    return False
        return True



from typing import List, Dict

from golem.structural_analysis.graph_sa.results.base_sa_approach_result import BaseSAApproachResult


class ReplaceSAApproachResult(BaseSAApproachResult):
    """ Class for presenting replacing result approaches. """
    def __init__(self):
        """ Main dictionary `self.metrics` contains entities as key and
        list with metrics as values"""
        self.metrics = dict()

    def add_results(self, entity_to_replace_to: str, metrics_values: List[float]):
        """ Sets value for specified metric. """
        self.metrics[entity_to_replace_to] = metrics_values

    def get_worst_result(self, metric_idx_to_optimize_by: int) -> float:
        """ Returns value of the worst metric. """
        return max([metrics[metric_idx_to_optimize_by] for metrics in list(self.metrics.values())])

    def get_worst_result_with_names(self, metric_idx_to_optimize_by: int) -> dict:
        """ Returns the worst metric among all calculated with its name and node's to replace to name. """
        worst_value = self.get_worst_result(metric_idx_to_optimize_by=metric_idx_to_optimize_by)
        for entity in self.metrics:
            if list(self.metrics[entity])[metric_idx_to_optimize_by] == worst_value:
                return {'value': worst_value, 'entity_to_replace_to': entity}

    def get_dict_results(self) -> Dict[int, List[float]]:
        """ Returns dict representation of results. """
        return self.metrics

    def get_rounded_metrics(self, idx: int = 2) -> dict:
        rounded = {}
        for metric in self.metrics:
            rounded[metric] = [round(metric, idx) for metric in self.metrics[metric]]
        return rounded

    def __str__(self):
        return 'ReplaceSAApproachResult'



from datetime import timedelta
from functools import partial
from typing import Type, Optional, Sequence

import networkx as nx

from examples.synthetic_graph_evolution.experiment_setup import run_experiments
from examples.synthetic_graph_evolution.generators import generate_labeled_graph
from golem.core.adapter.nx_adapter import BaseNetworkxAdapter
from golem.core.dag.verification_rules import DEFAULT_DAG_RULES
from golem.core.optimisers.genetic.gp_optimizer import EvoGraphOptimizer
from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters
from golem.core.optimisers.genetic.operators.base_mutations import MutationTypesEnum
from golem.core.optimisers.genetic.operators.crossover import CrossoverTypesEnum
from golem.core.optimisers.objective import Objective
from golem.core.optimisers.optimization_parameters import GraphRequirements
from golem.core.optimisers.optimizer import GraphGenerationParams, GraphOptimizer, AlgorithmParameters
from golem.metrics.edit_distance import tree_edit_dist
from golem.metrics.graph_metrics import degree_distance


def tree_search_setup(target_graph: Optional[nx.DiGraph] = None,
                      objective: Optional[Objective] = None,
                      optimizer_cls: Type[GraphOptimizer] = EvoGraphOptimizer,
                      algorithm_parameters: Optional[AlgorithmParameters] = None,
                      node_types: Sequence[str] = ('x',),
                      timeout: Optional[timedelta] = None,
                      num_iterations: Optional[int] = None):
    if target_graph is not None and objective is not None:
        raise ValueError('Please provide either target or objective, not both')
    elif target_graph is not None:
        # Setup objective:
        # - primary metric is edit distance between 2 trees
        # - secondary metric is difference in node degree distribution
        objective = Objective(
            quality_metrics={'edit_dist': partial(tree_edit_dist, target_graph)},
            complexity_metrics={'degree': partial(degree_distance, target_graph)},
            is_multi_objective=False,
        )
        max_graph_size = target_graph.number_of_nodes()
    elif objective is not None:
        max_graph_size = 1000
    else:
        raise ValueError()

    # Setup parameters
    requirements = GraphRequirements(
        max_arity=max_graph_size,
        max_depth=max_graph_size,
        early_stopping_timeout=10,
        early_stopping_iterations=num_iterations // 3 if num_iterations else None,
        keep_n_best=4,
        max_graph_fit_time=timedelta(seconds=10),
        timeout=timeout,
        num_of_generations=num_iterations,
        n_jobs=-1,
        history_dir=None,
    )
    default_gp_params = GPAlgorithmParameters(
        multi_objective=objective.is_multi_objective,
        mutation_types=[
            MutationTypesEnum.single_add,
            MutationTypesEnum.single_drop,
        ],
        crossover_types=[CrossoverTypesEnum.none]
    )
    gp_params = algorithm_parameters or default_gp_params
    graph_gen_params = GraphGenerationParams(
        adapter=BaseNetworkxAdapter(),
        rules_for_constraint=DEFAULT_DAG_RULES,
        available_node_types=node_types,
    )

    # Generate simple initial population with small tree graphs
    initial_graphs = [generate_labeled_graph('tree', 5, node_types)
                      for _ in range(gp_params.pop_size)]
    # Build the optimizer
    optimiser = optimizer_cls(objective, initial_graphs, requirements, graph_gen_params, gp_params)
    return optimiser, objective


if __name__ == '__main__':
    """
    In this experiment Optimizer is expected to find the target graph
    (exact or almost exact, depending on available time and graph complexity)
    using Tree Edit Distance metric and a random tree (nx.random_tree) as target.

    This convergence can be seen from achieved metrics and visually from graph plots.
    """
    results_log = run_experiments(optimizer_setup=tree_search_setup,
                                  optimizer_cls=EvoGraphOptimizer,
                                  graph_names=['tree'],
                                  graph_sizes=[16],
                                  num_trials=1,
                                  trial_timeout=5,
                                  trial_iterations=100,
                                  visualize=True)
    print(results_log)
